from builtins import object
import abc
import opentuner
from opentuner.resultsdb.models import *
from future.utils import with_metaclass


class InputManager(with_metaclass(abc.ABCMeta, object)):
  """
  abstract base class for compile and measurement
  """

  def set_driver(self, measurement_driver):
    self.driver = measurement_driver
    self.session = measurement_driver.session
    self.program = measurement_driver.tuning_run.program

  @abc.abstractmethod
  def select_input(self, desired_result):
    """
    select the input to be used to test desired_result
    """
    return opentuner.resultsdb.models.Input()


  def before_run(self, desired_result, input):
    """hook called before an input is used"""
    pass

  def after_run(self, desired_result, input):
    """hook called after an input is used"""
    pass

  def get_input_class(self):
    return None


class FixedInputManager(InputManager):
  """
  an input manage that produces a single input for all tests
  """

  def __init__(self,
               input_class_name='fixed',
               size=-1,
               path=None,
               extra=None):
    self.input_class_name = input_class_name
    self.size = size
    self.path = path
    self.extra = extra
    self.the_input = None
    super(FixedInputManager, self).__init__()


  def get_input_class(self):
    return InputClass.get(self.session,
                          program=self.program,
                          name=self.input_class_name,
                          size=self.size)

  def create_input(self, desired_result):
    """create the fixed input database object, result will be cached"""
    return Input(input_class=self.get_input_class(),
                 path=self.path,
                 extra=self.extra)

  def select_input(self, desired_result):
    if self.the_input is None:
      self.the_input = self.create_input(desired_result)
    return self.the_input









#!usr/bin/python

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from builtins import str
from builtins import zip
from builtins import map
from builtins import range
from past.utils import old_div
if __name__ == '__main__':
  from . import adddeps

import itertools
import math
import matplotlib.pyplot as plt
import numpy
import os
import sqlalchemy
import sqlalchemy.orm.exc

from collections import defaultdict
from fn import _
from fn import Stream
from fn.iters import repeat
from opentuner import resultsdb

PCTSTEPS = list(map(old_div(_,20.0), list(range(21))))


def mean(vals):
  """
  Arguments,
    vals: List of floating point numbers
  Returns,
    The mean of the numbers in the input list
    None if all values in the list are None
  """
  filtered_values = [float(x) for x in vals if x is not None]
  if len(filtered_values) == 0:
    return None
  return numpy.mean(numpy.array(filtered_values))


def stddev(vals):
  """
  Arguments,
    vals: List of floating point numbers
  Returns,
    The standard deviation of numbers in the input list
    None if all values in the list are None
  """
  filtered_values = [float(x) for x in vals if x is not None]
  if len(filtered_values) == 0:
    return None
  return math.sqrt(numpy.var(numpy.array(filtered_values)))


def get_dbs(path, db_type='sqlite:///'):
  """
  Arguments,
    path: Path of directory containing .db files
  Returns,
    A list of (engine, session) pairs to the dbs pointed to by
    the db files
  """
  dbs = list()
  for f in os.listdir(path):
    if 'journal' in f:
      continue
    try:
      db_path = os.path.join(path, f)
      e, sm = resultsdb.connect(db_type + db_path)
      dbs.append(sm())
    except Exception as e:
      print(e)
      print("Error encountered while connecting to db")
  return dbs


def matplotlibplot_file(labels, xlim = None, ylim = None, disp_types=['median']):
  """
  Arguments,
    labels: List of labels that need to be included in the plot
    xlim: Integer denoting the maximum X-coordinate in the plot
    ylim: Integer denoting the maximum Y-coordinate in the plot
    disp_types: List of measures that are to be displayed in the plot
  Returns,
    A figure object representing the required plot
  """

  figure = plt.figure()
  values = get_values(labels)
  for label in values:
    (mean_values, percentile_values) = values[label]
    for disp_type in disp_types:
      cols = None
      data = percentile_values

      if disp_type == 'median':
        cols = [11]
      elif disp_type == 'mean':
        cols = [1]
        data = mean_values
      elif disp_type == 'all_percentiles':
        cols = list(range(1,22))

      plotted_data = [[] for x in range(len(cols))]

      x_indices = []
      for data_point in data[1:]:
        x_indices.append(int(data_point[0]))
        for i in range(0, len(cols)):
          plotted_data[i].append(float(data_point[cols[i]]))
      args = []
      for to_plot in plotted_data:
        args.append(x_indices)
        args.append(to_plot)

      plt.plot(*args, label='%s(%s)' % (label, disp_type))

  if xlim is not None:
    plt.xlim(xlim)
  if ylim is not None:
    plt.ylim(ylim)

  plt.xlabel('Autotuning Time (seconds)')
  plt.ylabel('Execution Time (seconds)')
  plt.legend(loc='upper right')
  return figure


def run_label(tr):
  techniques = ','.join(tr.args.technique)
  if not tr.name or tr.name == 'unnamed':
    return techniques
  return tr.name


def combined_stats_over_time(label,
                             runs,
                             objective,
                             worst,
                             best,
                             ):
  """
  combine stats_over_time() vectors for multiple runs
  """

  extract_fn = _.result.time
  combine_fn = min
  no_data = 999

  by_run = [stats_over_time(session, run, extract_fn, combine_fn, no_data)
            for run, session in runs]
  max_len = max(list(map(len, by_run)))

  by_run_streams = [Stream() << x << repeat(x[-1], max_len-len(x))
                    for x in by_run]
  by_quanta = list(zip(*by_run_streams[:]))

  # TODO: Fix this, this variable should be configurable
  stats_quanta = 10
  def get_data(value_function):
    final_values = []
    for quanta, values in enumerate(by_quanta):
      sec = quanta*stats_quanta
      final_values.append([sec] + value_function(values))
    return final_values

  mean_values = get_data(lambda values: [mean(values), stddev(values)])

  def extract_percentiles(values):
    values = sorted(values)
    return ([values[int(round(p*(len(values)-1)))] for p in PCTSTEPS]
           + [mean(values)])
  percentile_values = get_data(extract_percentiles)
  return mean_values, percentile_values


def stats_over_time(session,
                    run,
                    extract_fn,
                    combine_fn,
                    no_data = None):
  """
  return reduce(combine_fn, map(extract_fn, data)) for each quanta of the
  tuning run
  """
  value_by_quanta = [ no_data ]
  start_date = run.start_date

  subq = (session.query(resultsdb.models.Result.id)
         .filter_by(tuning_run = run, was_new_best = True, state='OK'))

  q = (session.query(resultsdb.models.DesiredResult)
       .join(resultsdb.models.Result)
       .filter(resultsdb.models.DesiredResult.state=='COMPLETE',
               resultsdb.models.DesiredResult.tuning_run == run,
               resultsdb.models.DesiredResult.result_id.in_(subq.subquery()))
       .order_by(resultsdb.models.DesiredResult.request_date))

  first_id = None
  for dr in q:
    if first_id is None:
      first_id = dr.id
    td = (dr.request_date - start_date)
    duration = td.seconds + (td.days * 24 * 3600.0)
    # TODO: Make this variable configurable
    by_request_count = True
    stats_quanta = 10
    if by_request_count:
      quanta = dr.id - first_id
    else:
      quanta = int(old_div(duration, stats_quanta))
    while len(value_by_quanta) <= quanta:
      value_by_quanta.append(value_by_quanta[-1])

    if value_by_quanta[-1] is no_data:
      value_by_quanta[-1] = extract_fn(dr)
    else:
      value_by_quanta[-1] = combine_fn(value_by_quanta[-1], extract_fn(dr))

  return value_by_quanta


def get_all_labels():
  """
  Returns,
    List of labels that are in the complete state
  """
  dbs = get_dbs(os.getcwd())
  all_labels = list()
  for db in dbs:
    all_labels.extend(db.query(resultsdb.models.TuningRun.name)
                        .filter_by(state='COMPLETE')
                        .distinct()
                        .all())
  all_labels = [str(element[0]) for element in all_labels]
  return all_labels


def get_values(labels):
  """
  Arguments,
    labels: List of labels whose values are of interest
  Returns,
    A list of (mean, percentile) tuples, corresponding to the
    provided list of labels
  """
  dbs = get_dbs(os.getcwd())
  dir_label_runs = defaultdict(lambda: defaultdict(list))
  for db in dbs:
    q = (db.query(resultsdb.models.TuningRun)
            .filter_by(state='COMPLETE')
            .order_by('name'))
    if labels:
      q = q.filter(resultsdb.models.TuningRun.name.in_(labels))
    for tr in q:
      dir_label_runs[run_label(tr)][run_label(tr)].append((tr, db))
  all_run_ids = list()
  returned_values = {}
  for d, label_runs in list(dir_label_runs.items()):
    all_run_ids = list(map(_[0].id, itertools.chain(*list(label_runs.values()))))
    session = list(label_runs.values())[0][0][1]
    objective = list(label_runs.values())[0][0][0].objective

    q = (session.query(resultsdb.models.Result)
         .filter(resultsdb.models.Result.tuning_run_id.in_(all_run_ids))
         .filter(resultsdb.models.Result.time < float('inf'))
         .filter_by(was_new_best=True, state='OK'))
    total = q.count()
    q = objective.filter_acceptable(q)
    acceptable = q.count()
    q = q.order_by(*objective.result_order_by_terms())
    best = q.limit(1).one()
    worst = q.offset(acceptable - 1).limit(1).one()

    for label, runs in sorted(label_runs.items()):
      (mean_values, percentile_values) = combined_stats_over_time(label, runs, objective, worst, best)
      returned_values[label] = (mean_values, percentile_values)
      final_scores = list()
      for run, session in runs:
        try:
          final = (session.query(resultsdb.models.Result)
                  .filter_by(tuning_run = run,
                             configuration = run.final_config)
                  .limit(1).one())
        except sqlalchemy.orm.exc.NoResultFound:
          continue
        final_scores.append(objective.stats_quality_score(final, worst, best))
      final_scores.sort()
  return returned_values

if __name__ == '__main__':
    labels = [u'timeouts', u'always_reorder', u'add_store_at', u'all_options']
    get_values(labels)
    print(get_all_labels())



# -*- coding: utf-8 -*-
# vim: tabstop=2 shiftwidth=2 softtabstop=2 expandtab autoindent smarttab
from __future__ import absolute_import
from builtins import range
from builtins import object
from .manipulator import *
from opentuner.search import technique
import random
import math

class PSO(technique.SequentialSearchTechnique ):
  """ Particle Swarm Optimization """
  def __init__(self, crossover, N = 30, init_pop=None, *pargs, **kwargs):
    """
    crossover: name of crossover operator function
    """
    super(PSO, self).__init__(*pargs, **kwargs)
    self.crossover = crossover
    self.name = 'pso-'+crossover.replace("op3_cross_","")
    self.init_pop = init_pop
    self.N = N

  def main_generator(self):

    objective   = self.objective
    driver    = self.driver
    m = self.manipulator
    def config(cfg):
      return driver.get_configuration(cfg)
  
    population = self.init_pop
    if not population:
      population = [HybridParticle(m, self.crossover) for i in range(self.N)]

    for p in population:
      yield driver.get_configuration(p.position)

    while True:
      for particle in population:
        g = driver.best_result.configuration.data
        old=m.copy(particle.position)
        particle.move(g)
        yield config(particle.position)
        # update individual best
        if objective.lt(config(particle.position), config(particle.best)):
          particle.best = particle.position

class HybridParticle(object):
  def __init__(self, m, crossover_choice, omega=0.5, phi_l=0.5, phi_g=0.5):

    """
    m: a configuraiton manipulator
    omega: influence of the particle's last velocity, a float in range [0,1] ; omega=1 means even speed
    phi_l: influence of the particle's distance to its historial best position, a float in range [0,1]
    phi_g: influence of the particle's distance to the global best position, a float in range [0,1]
    """

    self.manipulator = m
    self.position = self.manipulator.random()   
    self.best = self.position
    self.omega = omega
    self.phi_l = phi_l
    self.phi_g = phi_g
    self.crossover_choice = crossover_choice
    self.velocity = {}
    for p in self.manipulator.params:
      # Velocity as a continous value
      self.velocity[p.name]=0  

  def move(self, global_best):
    """
    Update parameter values using corresponding operators. 
    TODO: introduce operator choice map
    """
    m = self.manipulator
    for p in m.params:
      self.velocity[p.name] = p.op3_swarm(self.position, global_best, self.best, c=self.omega, c1=self.phi_g, c2=self.phi_l, xchoice=self.crossover_choice, velocity=self.velocity[p.name])


technique.register(PSO(crossover = 'op3_cross_OX3'))
technique.register(PSO(crossover = 'op3_cross_OX1'))
technique.register(PSO(crossover = 'op3_cross_PMX'))
technique.register(PSO(crossover = 'op3_cross_PX'))
technique.register(PSO(crossover = 'op3_cross_CX'))



from __future__ import print_function
# vim: tabstop=2 shiftwidth=2 softtabstop=2 expandtab autoindent smarttab
from builtins import object
import argparse
import copy
import inspect
import logging
import math
import os
import socket
import sys
import time
import uuid
from datetime import datetime

from opentuner import resultsdb
from opentuner.search.driver import SearchDriver
from opentuner.measurement.driver import MeasurementDriver

log = logging.getLogger(__name__)

argparser = argparse.ArgumentParser(add_help=False)
argparser.add_argument('--label',
                       help="name for the TuningRun")
argparser.add_argument('--print-search-space-size', action='store_true',
                       help="Print out the estimated size of the search space and exit")
argparser.add_argument('--database',
                       help=("database to store tuning results in, see: "
                             "http://docs.sqlalchemy.org/en/rel_0_8/core/engines.html#database-urls"))
argparser.add_argument('--print-params','-pp',action='store_true',
                       help='show parameters of the configuration being tuned')


class CleanStop(Exception):
  pass


class LogFormatter(logging.Formatter):
  def format(self, record):
    record.relativeCreated /= 1000.0
    try:
      # python 2.7
      return super(LogFormatter, self).format(record)
    except:
      # python 2.6
      return _OldFormatter.format(self, record)


_OldFormatter = logging.Formatter
logging.Formatter = LogFormatter

try:
  # python 2.7
  from logging.config import dictConfig
except:
  # python 2.6
  from .utils.dictconfig import dictConfig

the_logging_config = {
  'version': 1,
  'disable_existing_loggers': False,
  'formatters': {'console': {'format': '[%(relativeCreated)6.0fs] '
                                       '%(levelname)7s %(name)s: '
                                       '%(message)s'},
                 'file': {'format': '[%(asctime)-15s] '
                                    '%(levelname)7s %(name)s: '
                                    '%(message)s '
                                    '@%(filename)s:%(lineno)d'}},
  'handlers': {'console': {'class': 'logging.StreamHandler',
                           'formatter': 'console',
                           'level': 'INFO'},
               'file': {'class': 'logging.FileHandler',
                        'filename': 'opentuner.log',
                        'formatter': 'file',
                        'level': 'WARNING'}},
  'loggers': {'': {'handlers': ['console', 'file'],
                   'level': 'INFO',
                   'propagate': True}}}


def init_logging():
  dictConfig(the_logging_config)
  global init_logging
  init_logging = lambda: None


class TuningRunMain(object):
  def __init__(self,
               measurement_interface,
               args,
               search_driver=SearchDriver,
               measurement_driver=MeasurementDriver):
    init_logging()

    manipulator = measurement_interface.manipulator()
    if args.print_search_space_size:
      print("10^{%.2f}" % math.log(manipulator.search_space_size(), 10))
      sys.exit(0)
    # show internal parameter representation
    if args.print_params:
      cfg = manipulator.seed_config()
      d = manipulator.parameters_dict(cfg)
      params_dict ={} 
      for k in d: 
        cls = d[k].__class__.__name__
        p = (k, d[k].search_space_size())
        if cls in params_dict:
          params_dict[cls].append(p)
        else:
          params_dict[cls] = [p]
      for k in params_dict:
        print(k, params_dict[k])
        print()
      sys.exit(0)

    input_manager = measurement_interface.input_manager()
    objective = measurement_interface.objective()

    if not args.database:
      #args.database = 'sqlite://' #in memory
      if not os.path.isdir('opentuner.db'):
        os.mkdir('opentuner.db')
      args.database = 'sqlite:///' + os.path.join('opentuner.db',
                                                  socket.gethostname() + '.db')

    if '://' not in args.database:
      args.database = 'sqlite:///' + args.database

    if not args.label:
      args.label = 'unnamed'

    #self.fake_commit = ('sqlite' in args.database)
    self.fake_commit = True

    self.args = args

    self.engine, self.Session = resultsdb.connect(args.database)
    self.session = self.Session()
    self.tuning_run = None
    self.search_driver_cls = search_driver
    self.measurement_driver_cls = measurement_driver
    self.measurement_interface = measurement_interface
    self.input_manager = input_manager
    self.manipulator = manipulator
    self.objective = objective
    self.objective_copy = copy.copy(objective)
    self.last_commit_time = time.time()

  def init(self):
    if self.tuning_run is None:
      program_version = (self.measurement_interface
                         .db_program_version(self.session))
      self.session.flush()
      self.measurement_interface.prefix_hook(self.session)
      self.tuning_run = (
        resultsdb.models.TuningRun(
          uuid=uuid.uuid4().hex,
          name=self.args.label,
          args=self.args,
          start_date=datetime.now(),
          program_version=program_version,
          objective=self.objective_copy,
        ))
      self.session.add(self.tuning_run)

      driver_kwargs = {
        'args': self.args,
        'input_manager': self.input_manager,
        'manipulator': self.manipulator,
        'measurement_interface': self.measurement_interface,
        'objective': self.objective,
        'session': self.session,
        'tuning_run_main': self,
        'tuning_run': self.tuning_run,
        'extra_seeds': self.measurement_interface.seed_configurations(),
        'extra_criteria': self.measurement_interface.extra_convergence_criteria
      }

      self.search_driver = self.search_driver_cls(**driver_kwargs)

      self.measurement_driver = self.measurement_driver_cls(**driver_kwargs)
      self.measurement_interface.set_driver(self.measurement_driver)
      self.input_manager.set_driver(self.measurement_driver)

      self.tuning_run.machine_class = self.measurement_driver.get_machine_class()
      self.tuning_run.input_class = self.input_manager.get_input_class()

  def commit(self, force=False):
    if (force or not self.fake_commit or
            time.time() - self.last_commit_time > 30):
      self.session.commit()
      self.last_commit_time = time.time()
    else:
      self.session.flush()

  def main(self):
    self.init()
    try:
      self.tuning_run.state = 'RUNNING'
      self.commit(force=True)
      self.search_driver.main()
      if self.search_driver.best_result:
        self.measurement_interface.save_final_config(
            self.search_driver.best_result.configuration)
      self.tuning_run.final_config = self.search_driver.best_result.configuration
      self.tuning_run.state = 'COMPLETE'
    except:
      self.tuning_run.state = 'ABORTED'
      raise
    finally:
      self.tuning_run.end_date = datetime.now()
      self.commit(force=True)
      self.session.close()

  def results_wait(self, generation):
    """called by search_driver to wait for results"""
    #single process version:
    self.measurement_interface.pre_process()  
    self.measurement_driver.process_all()
    self.measurement_interface.post_process()

def main(interface, args, *pargs, **kwargs):
  if inspect.isclass(interface):
    interface = interface(args=args, *pargs, **kwargs)
  return TuningRunMain(interface, args).main()




from __future__ import print_function
from __future__ import absolute_import
from builtins import next
from builtins import map
from builtins import str
from builtins import object
import abc
import argparse
import logging
import os
import random
import sys

from importlib import import_module
from datetime import datetime
from fn import _

from opentuner.resultsdb.models import *
from .plugin import SearchPlugin
from future.utils import with_metaclass

log = logging.getLogger(__name__)
#log.setLevel(logging.DEBUG)

argparser = argparse.ArgumentParser(add_help=False)
argparser.add_argument('--technique','-t', action='append',
                       help="which technique to use")
argparser.add_argument('--list-techniques','-lt', action='store_true',
                       help="list techniques available and exit")
argparser.add_argument('--generate-bandit-technique','-gbt', action='store_true',
                       help="randomly generate a bandit to use")

class SearchTechniqueBase(with_metaclass(abc.ABCMeta, object)):
  """
  abstract base class for search techniques, with minimal interface
  """

  def __init__(self, name = None):
    super(SearchTechniqueBase, self).__init__()
    if name:
      self.name = name
    else:
      self.name = self.default_name()

  def is_ready(self):
    """test if enough data has been gathered to use this technique"""
    return True

  def default_name(self):
    """name of this SearchTechnique uses for display/accounting"""
    return self.__class__.__name__

  def handle_requested_result(self, result):
    """called for each new Result(), requested by this technique"""
    pass

  @abc.abstractmethod
  def set_driver(self, driver):
    """called at start of tuning process"""
    return

  @abc.abstractmethod
  def desired_result(self):
    """
    return at most count resultsdb.models.DesiredResult objects based on past
    performance
    """
    return

class SearchTechnique(SearchPlugin, SearchTechniqueBase):
  """
  a search search technique with basic utility functions
  """

  def __init__(self, *pargs, **kwargs):
    super(SearchTechnique, self).__init__(*pargs, **kwargs)
    self.driver = None
    self.manipulator = None
    self.objective = None
    self.request_count = 0

  def set_driver(self, driver):
    super(SearchTechnique, self).set_driver(driver)
    self.manipulator = driver.manipulator
    self.objective = driver.objective
    driver.add_plugin(self)

  def desired_result(self):
    """
    create and return a resultsdb.models.DesiredResult
    returns None if no desired results and False if waiting for results
    """
    cfg = self.desired_configuration()
    if cfg is None:
      return None
    if cfg is False:
      return False
    if type(cfg) is Configuration:
      config = cfg
    else:
      config = self.driver.get_configuration(cfg)
    desired = DesiredResult(configuration=config,
                            requestor=self.name,
                            generation=self.driver.generation,
                            request_date=datetime.now(),
                            tuning_run=self.driver.tuning_run)
    if hasattr(self, 'limit'):
      desired.limit = self.limit
    self.driver.register_result_callback(desired, self.handle_requested_result)
    self.request_count += 1
    return desired

  @abc.abstractmethod
  def desired_configuration(self):
    """
    return a cfg that we should test
    given a ConfigurationManipulator and SearchDriver
    return None if there are no configurations to test
    return False if waiting for results
    """
    return dict()

  def handle_requested_result(self, result):
    """called for each new Result(), regardless of who requested it"""
    pass

  def default_generated_name(self):
    """ The default generated name for this technique """
    return self.base_name()

  def use_default_generated_name(self):
    """ set the name of this technique to the default generated name """
    self.name = self.default_generated_name()

  def base_name(self):
    """
    Return the base name of this technique with form
    classname;hyperparam1,v1;hyperparam2,v2 ...
    where hyperparams are taken in order from get_hyper_parameters()

    Should only be called after this technique has finished initializing.
    """
    out = [self.__class__.__name__]
    for hyper_parameter in self.get_hyper_parameters():
      # get hyperparam,v as a string and append
      try:
        out.append(hyper_parameter + ',' + str(getattr(self, hyper_parameter)))
      except AttributeError:
        log.error("Uninitialized hyper-parameter %s for technique %s.",
                   hyper_parameter, self.__class__.__name__)

    return ';'.join(out)

  @classmethod
  def get_hyper_parameters(cls):
    """
    return a list of hyper-parameters names for this technique

    Name strings must match the corresponding attribute with the hyper-parameter
    value on technique instances. Names should also match the key word argument
    used when initializing an instance. Hyperparameters should only take literal
    values.

    For example, given hyper parameter "mutation_rate", then the __init__ method
    should have 'mutation_rate' as a key word argument and later have the line
    self.mutation_rate = mutation_rate
    """
    return []

  @classmethod
  def generate_technique(cls, manipulator=None, *args, **kwargs):
    """ return a new technique based off this instance """
    t = cls(*args, **kwargs)
    t.use_default_generated_name()
    return t

class PureRandom(SearchTechnique):
  """
  request configurations completely randomly
  """
  def desired_configuration(self):
    return self.manipulator.random()

class AsyncProceduralSearchTechnique(SearchTechnique):
  def __init__(self, *pargs, **kwargs):
    super(AsyncProceduralSearchTechnique, self).__init__(*pargs, **kwargs)
    self.gen = None
    self.done = False
    self.latest_results = []

  def call_main_generator(self):
    """passthrough (used in subclasses)"""
    return self.main_generator()

  def desired_configuration(self):
    if self.gen is None:
      log.debug("%s: creating generator", self.name)
      self.gen = self.call_main_generator()
    if not self.done:
      try:
        return next(self.gen)
      except StopIteration:
        log.debug("%s: generator finished", self.name)
        self.done = True
    return None

  @abc.abstractmethod
  def main_generator(self):
    """
    custom generator to conduct this search, should:
    yield config
    to request tests and call driver.get_results() to read the results

    in AsyncProceduralSearchTechnique results are ready at an undefined
    time (`yield False` to stall and wait for them)

    in SequentialSearchTechnique results are ready after the yield
    """
    pass

  def is_ready(self):
    return not self.done

class SequentialSearchTechnique(AsyncProceduralSearchTechnique):
  def __init__(self, novelty_threshold=50, reset_threshold=500, *pargs, **kwargs):
    super(SequentialSearchTechnique, self).__init__(*pargs, **kwargs)
    self.pending_tests = []
    self.novelty_threshold = novelty_threshold
    self.rounds_since_novel_request = 0
    self.reset_threshold = reset_threshold

  def yield_nonblocking(self, cfg):
    """
    within self.main_generator() act like `yield cfg`, but don't wait for the
    results until the following yield (spawn/sync style)
    """
    if cfg:
      self.pending_tests.append(cfg)

  def call_main_generator(self):
    """insert waits for results after every yielded item"""
    subgen = self.main_generator()
    self.rounds_since_novel_request = 0
    while True:
      self.rounds_since_novel_request += 1
      if (self.rounds_since_novel_request % self.novelty_threshold) == 0:
        log.warning("%s has not requested a new result for %d rounds",
                    self.name, self.rounds_since_novel_request)
        if (self.rounds_since_novel_request > self.reset_threshold):
          log.warning("%s is being reset", self.name)
          subgen = self.main_generator()
          self.rounds_since_novel_request = 0
        yield None # give other techniques a shot
      try:
        p = next(subgen)
        if p:
          self.pending_tests.append(p)
      except StopIteration:
        return
      finally:
        for p in self.pending_tests:
          if not self.driver.has_results(p):
            self.rounds_since_novel_request = 0
            yield p

      # wait for all pending_tests to have results
      c = 0
      while self.pending_tests:
        log.debug("%s: waiting for %d pending tests",
                  self.name, len(self.pending_tests))
        c += 1
        if (c % 100) == 0:
          log.error("%s: still waiting for %d pending tests (c=%d)",
                     self.name, len(self.pending_tests), c)

        self.pending_tests = [x for x in self.pending_tests if not self.driver.has_results(x)]
        if self.pending_tests:
          self.rounds_since_novel_request = 0
          yield False # wait

#list of all techniques
the_registry = list()

#list of technique generators
the_generator_registry = list()

def register(t):
  the_registry.append(t)

def register_generator(cls, generator_weight=1.0, *args, **kwargs):
  """
  register a technique generator - a tuple of (technique class, args, kwargs)
  where args and kwargs will be passed into the generate_technique classmethod -
  with specified probability weight when randomly choosing a generator

  :param cls: a technique class to use as a generator
  :param generator_weight: probability weighting when randomly choosing a generator
  :param args: arguments to pass into generate_technique class method
  :param kwargs: arguments to pass into generate_technique class method
  """
  the_generator_registry.append(((cls, args, kwargs), generator_weight))

register(PureRandom())

def get_random_generator_technique(generators=None, manipulator=None):
  """
  Takes in a sequence of ((generator, args, kwargs), weight) tuples.
  Returns a random generated technique info tuple

  :param generators: optional argument to avoid repeated getting of generators
  :param manipulator: manipulator to pass to generate_technique class method.
  """
  if generators is None:
    techniques, generators = all_techniques()
  g, args, kwargs = weighted_choice(generators)
  return g.generate_technique(manipulator, *args, **kwargs)


def weighted_choice(choices):
  """ takes in a sequence of (choice, weight) tuples and randomly returns one """
  total = sum(w for c, w in choices)
  r = random.uniform(0, total)
  upto = 0
  for c, w in choices:
    upto += w
    if upto > r:
      return c
  return random.choice([c for c, w in choices])


def all_techniques():
  #import all modules in search to ensure techniques are Registered
  for f in sorted(os.listdir(os.path.dirname(__file__))):
    m = re.match(r'^(.*)[.]py$', f)
    if m:
      import_module('opentuner.search.'+m.group(1))

  return the_registry, the_generator_registry

def get_enabled(args):
  techniques, generators = all_techniques()
  if args.list_techniques:
    for t in techniques:
      print(t.name)
    sys.exit(0)

  if not args.technique:
    # no techniques specified, default technique
    args.technique = ['AUCBanditMetaTechniqueA']

  for unknown in set(args.technique) - set(map(_.name, techniques)):
    log.error('unknown technique %s', unknown)
    raise Exception('Unknown technique: --technique={}'.format(unknown))

  return [t for t in techniques if t.name in args.technique]

def get_root(args):
  from .metatechniques import RoundRobinMetaSearchTechnique
  enabled = get_enabled(args)
  if len(enabled) == 1:
    return enabled[0]
  return RoundRobinMetaSearchTechnique(get_enabled(args))




from __future__ import division
from __future__ import absolute_import
from builtins import map
from builtins import range
from past.utils import old_div
from builtins import object
import random
import time
import logging
from fn import _
from .technique import register
from .technique import SearchTechnique

log = logging.getLogger(__name__)
log.setLevel(logging.WARNING)


class PopulationMember(object):
  def __init__(self, config, submitted=True):
    self.config = config
    self.submitted = submitted
    self.timestamp = time.time()
    self.candidate_replacement = None

  def touch(self):
    self.timestamp = time.time()


class DifferentialEvolution(SearchTechnique):
  """
  based on http://cci.lbl.gov/cctbx_sources/scitbx/differential_evolution.py
  """

  def __init__(self,
               population_size=30,
               cr=0.9,  # crossover rate
               n_cross=1,  # force at least 1 to crossover
               information_sharing=1,  # number token sharing pop members
               duplicate_retries=5,  # how many times to retry on duplicate
               *pargs, **kwargs):

    self.population_size = population_size
    self.cr = cr
    self.n_cross = n_cross
    self.information_sharing = information_sharing
    self.population = None
    self.duplicate_retries = duplicate_retries
    self.limit = None
    super(DifferentialEvolution, self).__init__(*pargs, **kwargs)

  @classmethod
  def get_hyper_parameters(cls):
    return ['population_size', 'cr', 'n_cross', 'information_sharing']

  def initial_population(self):
    self.population = [PopulationMember(
        self.driver.get_configuration(
            self.manipulator.random()), submitted=False)
        for z in range(self.population_size)]

  def oldest_pop_member(self):
    # since tests are run in parallel, exclude things with a replacement pending
    pop_without_replacements = [x for x in self.population if x.candidate_replacement is None]
    if not pop_without_replacements:
      # everything has a pending replacement
      return None
    pop_without_replacements.sort(key=_.timestamp)
    return pop_without_replacements[0]

  def desired_configuration(self):
    """
    return a cfg that we should test,
    """
    if not self.population:
      # first time called
      self.initial_population()

    # make sure initial population is completely submitted
    for p in self.population:
      if not p.submitted:
        p.submitted = True
        if p is self.population[-1]:
          log.info('initial population testing done')
        return p.config

    # pp is member of population to be replaced
    oldest_pop_member = self.oldest_pop_member()
    if not oldest_pop_member:
      return None

    config = None
    for retry in range(self.duplicate_retries):
      config = self.driver.get_configuration(
          self.create_new_configuration(oldest_pop_member))
      if not self.driver.has_results(config):
        break
      # new configuration would have been a duplicate, try again

    oldest_pop_member.touch()  # move to back of the line for next replacement
    oldest_pop_member.candidate_replacement = config
    self.limit = self.driver.objective.limit_from_config(
        oldest_pop_member.config)
    return oldest_pop_member.candidate_replacement

  def create_new_configuration(self, parent_pop_member):
    cfg = self.manipulator.copy(parent_pop_member.config.data)
    cfg_params = self.manipulator.proxy(cfg)

    # pick 3 random parents, not pp
    shuffled_pop = list(set(self.population) - set([parent_pop_member]))

    # share information with other techniques
    if self.driver.best_result:
      shuffled_pop += ([PopulationMember(self.driver.best_result.configuration)]
                       * self.information_sharing)

    random.shuffle(shuffled_pop)
    x1, x2, x3 = list(map(_.config.data, shuffled_pop[0:3]))

    use_f = old_div(random.random(), 2.0) + 0.5

    params = self.manipulator.param_names(cfg, x1, x2, x3)
    random.shuffle(params)
    for i, k in enumerate(params):
      if i < self.n_cross or random.random() < self.cr:
        # cfg = x1 + use_f*(x2 - x3)
        cfg_params[k].op4_set_linear(x1, x2, x3, 1.0, use_f, -use_f)

    return cfg

  def handle_requested_result(self, result):
    """called when new results are added"""
    for p in self.population:
      if p.candidate_replacement == result.configuration:
        if self.objective.lt(p.candidate_replacement, p.config):
          # candidate replacement was better, replace it!
          p.config = p.candidate_replacement
          log.info('better point')
        p.candidate_replacement = None


class DifferentialEvolutionAlt(DifferentialEvolution):
  def __init__(self, cr=0.2, **kwargs):
    kwargs['cr'] = cr
    super(DifferentialEvolutionAlt, self).__init__(**kwargs)


register(DifferentialEvolution())
register(DifferentialEvolutionAlt())
register(DifferentialEvolution(population_size=100, cr=0.2,
                               name='DifferentialEvolution_20_100'))





from builtins import object
from opentuner.resultsdb.models import *


class DriverBase(object):
  """
  shared base class between MeasurementDriver and SearchDriver
  """

  def __init__(self,
               session,
               tuning_run,
               objective,
               tuning_run_main,
               args,
               **kwargs):
    self.args = args
    self.objective = objective
    self.session = session
    self.tuning_run_main = tuning_run_main
    self.tuning_run = tuning_run
    self.program = tuning_run.program

  def results_query(self,
                    generation=None,
                    objective_ordered=False,
                    config=None):
    q = self.session.query(Result)
    q = q.filter_by(tuning_run=self.tuning_run)

    if config:
      q = q.filter_by(configuration=config)

    if generation is not None:
      subq = (self.session.query(DesiredResult.result_id)
              .filter_by(tuning_run=self.tuning_run,
                         generation=generation))
      q = q.filter(Result.id.in_(subq.subquery()))

    if objective_ordered:
      q = self.objective.result_order_by(q)

    return q

  def requests_query(self):
    q = self.session.query(DesiredResult).filter_by(tuning_run=self.tuning_run)
    return q
    




from __future__ import absolute_import
from past.builtins import cmp
from builtins import map
from builtins import range
from builtins import object
import random
import time
import sys
import json
from fn import _
from .technique import all_techniques
from .technique import register
from .technique import register_generator
from .technique import SequentialSearchTechnique
from .manipulator import *
from opentuner.search.manipulator import Parameter
from future.utils import with_metaclass
from functools import reduce


class PopulationMember(object):
  """
  An extendable object representing a population member for ComposableEvolutionaryTechniques.
  Must have the field "config" which is a configuration
  """
  def __init__(self, config):
    self.config = config
    self.timestamp = time.time()

  def touch(self):
    """
    Update the timestamp on a PopulationMember
    """
    self.timestamp = time.time()


class ComposableEvolutionaryTechnique(with_metaclass(abc.ABCMeta, SequentialSearchTechnique)):
  """
  An abstract base class for a technique that is composable with operators
  """

  # operator_map - from param-type to dict with operator name + list of arguments TODO
  # min_parent - minimum number of parents returned. Limits which operators can be used
  def __init__(self,
               operator_map = {},
               population_size = 50,
               initial_configs = None,
               *pargs,
               **kwargs):
    """

    :param operator_map:
    :param population_size:
    :param initial_configs:
    :param pargs:
    :param kwargs:
    :return:
    """
    super(ComposableEvolutionaryTechnique, self).__init__(*pargs, **kwargs)
    # generate a name based on operators if no name

    self.initial_configurations = initial_configs
    self.population_size = population_size
    self.operator_map = operator_map # map from parameter type to an operator function

  def set_operator_map(self, operator_map):
    self.operator_map = operator_map

  @classmethod
  def get_hyper_parameters(cls):
    return ['population_size']

  def default_generated_name(self):
    """
    Gets the default name for this technique based on its operator map

    Name is in the format
    classname paramname;opname;[arg1,arg2,[[kwarg1,v1][kwarg2,v2]]] paramname2;opname2;...
    """
    # TODO - include technique hyperparameters
    parts = [self.base_name()]
    for param in sorted(self.operator_map, cmp=lambda x,y: cmp(x.__name__, y.__name__)):
      subparts = [param.__name__]
      operator_info = self.operator_map[param]
      subparts.append(operator_info['op_name'])
      args = list(operator_info['args'])
      kwargs = operator_info['kwargs']
      args.append([(k,kwargs[k]) for k in sorted(kwargs)])
      subparts.append(json.dumps(args, separators=(',', ':')))
      parts.append(';'.join(subparts))
    return ' '.join(parts)


  def make_population_member(self, config):
    """
    Given a configuration, returns an object representing a single member of the
    population with the given configuration. Meta-data about the configuration,
    such as last selection time as a parent, can be attached to the object.

    This can be overriden to return a custom population member for use in
    :py:meth:`get_parents` and :py:meth:`update_population`

    :param config: the configuration that this population member will represent
    :return: a population member reresenting the input configuration.
    """
    return PopulationMember(config)

  def select_parameters(self, params):
    """
    Given all the available parameters, return a subset of parameters to operate
    on when generating a new configuration.

    Can override this to operate on only a subset of parameters.

    :param params: a list of all the available parameters
    :return: a subset of params
    """
    return params

  @abc.abstractmethod
  def minimum_number_of_parents(self):
    """
    Return the minimum number of parents ever returned by :py:meth:`get_parents`.
    This limits which operators can be composed with the technique. Operators
    requiring more input configurations than the minimum number of parents will
    result in an error.

    :return: the minimum number of parents ever generated.
    """
    return 1

  @abc.abstractmethod
  def get_parents(self, population):
    """
    Given the current population, return a list of configurations that will be
    used to generate a new configuration via operators. Returning less parents
    than guaranteed by :py:meth:`minimum_number_of_parents` results in an error.

    The parents will be passed to operators in order. If there are more parents
    than required by an operator, extra parents will be discarded.

    Note that operators mutate the first configuration passed in.

    :param population: the current population in the technique
    :return: a list of parent configurations to generate a new configuration from
    """
    pass

  @abc.abstractmethod
  def update_population(self, config, population):
    """
    Update the population given the newest configuration and current population
    in the technique. should return the new population

    :param config: the newest generated configuration
    :param population: the current population in this iteration of the technique
    :return: the updated population
    """
    pass

  def get_initial_population(self):
    """
    Returns an initial population by passing initial configurations into
    :py:meth:`make_population_member`

    :return: an initial list of objects returned by :py:meth:`make_population_member`.
    """
    init_configs = self.initial_configurations
    if not init_configs:
      init_configs = [self.manipulator.random() for i in range(self.population_size)]
    return [PopulationMember(config) for config in init_configs]

  def lt(self, cfg_a, cfg_b):
    """
    Return whether cfg_a has a better objective function score than cfg_b

    :param cfg_a: first configuration
    :param cfg_b: second configuration
    :return: True if cfg_a is better than cfg_b
    """
    def config(cfg):
      return self.driver.get_configuration(cfg)
    return self.objective.lt(config(cfg_a), config(cfg_b))

  def lte(self, cfg_a, cfg_b):
    """
    Return whether cfg_a's objective function score is at least as good as cfg_b's
    score

    :param cfg_a: first configuration
    :param cfg_b: second configuration
    :return: True if cfg_a is at least as good as cfg_b
    """
    def config(cfg):
      return self.driver.get_configuration(cfg)
    return self.objective.lte(config(cfg_a), config(cfg_b))

  def get_global_best_configuration(self):
    """
    Return the current global best configuration in the search

    :return: the current global best configuration
    """
    if (self.driver.best_result is not None and
        self.driver.best_result.state == 'OK'):
      return self.manipulator.copy(self.driver.best_result.configuration.data)
    else:
      return self.manipulator.random()

  def get_default_operator(self, param_type):
    """
    Given a parameter type, return a dictionary with information about the
    operator to be used for the parameter. The returned dictionary must contain
    the following 3 key, value pairs

      1. 'op_name' - the string name of the operator
      2. 'args' - an iterable of the non-configuration arguments in order
      3. 'kwargs' - a dictionary from any optional arguments to their values

    :return: a dictionary containing information about the operator to apply for the input parameter type
    """
    return {'op_name': 'op1_nop', 'args': [], 'kwargs': {}}

  # HELPER METHODS FOR BUILDING OPERATOR MAP
  @classmethod
  def add_to_map(cls, operator_map, param_type, operator_name, *args, **kwargs):
    """
    A helper method for adding parameter to operator mappings into the operator
    map.

    :param operator_map: the operator map to add to
    :param param_type: the parameter type to use the this operator on
    :param operator_name: the string name of the operator method
    :param *args: any non-configuration arguments to the operator
    :param **kwargs: any keyword arguemnts for the operator
    """
    if(isinstance(param_type, Parameter)):
      ptype = type(param_type)
    elif (type(param_type) == str):
      ptype = reduce(getattr, param_type.split("."), sys.modules[__name__])
    else:
      ptype = param_type;

    operator_map[ptype] = {'op_name': operator_name, 'args':args, 'kwargs':kwargs}


  def main_generator(self):
    """
    The primary body of the search technique.
    Initializes an initial population and then yields configurations by applying
    operators to get_parents.
    """
    min_parents = self.minimum_number_of_parents();
    # convert a manipulator configuration to a db.models.Configuration
    def get_driver_configuration(cfg):
      return self.driver.get_configuration(cfg)

    # initialize the population
    population = self.get_initial_population()

    # measure initial population
    for p in population:
      yield get_driver_configuration(p.config)

    while True:
      # get parents
      parents = self.get_parents(population)
      if len(parents) < min_parents:
         log.error("%s: Number of parents returned %d is less than the guaranteed"
                     + " minimum returned by minimum_number_of_parents() %d. ",
                     self.name, len(parents), min_parents)
         # fail and let other techniques work forever
         while True:
          yield None

      params = self.select_parameters(self.manipulator.params)
      config = self.get_new_config(parents, params)
      yield get_driver_configuration(config)

      population = self.update_population(config, population)

      # safety check that population has all been tested
      for p in population:
        if not self.driver.has_results(get_driver_configuration(p.config)):
          yield get_driver_configuration(p.config)

  def get_new_config(self, parents, params):
    """
    Return a new configuration to test, given a list of parent configurations
    This mutates the first parent

    :param parents: A list of parent configurations
    :params: A list of parameters to operate on
    :return: The mutated configuration (first parent)
    """
    for param in params:
      self.apply_operator(param, parents) #TODO
    return parents[0]

  def apply_operator(self, param, parents):
    """
    Apply the appropriate operator for param to parents.
    If an operator takes less input configurations than the number of parents,
    only the first parents are passed in. If operator takes more input configs
    than minimum_number_of_parents, logs an error and doesn't do anything
    """
    x = self.get_operator(type(param))

    operator_name = x['op_name']
    if not self.is_valid_operator(type(param), operator_name):
      # do nothing
      return

    # operator is already in valid form and starts with op1, op2, op3, op4, or opn
    num_parents_required = operator_name[2]
    if num_parents_required == 'n':
      args = parents[0] + [parents[1:]]
    else:
      num_parents_required = int(num_parents_required)
      args = parents[:num_parents_required]
    args.extend(x['args'])

    kwargs = x['kwargs']

    getattr(param, operator_name)(*args, **kwargs)

  def get_operator(self, param_type):
    if param_type in self.operator_map:
      return self.operator_map[param_type]
    return self.get_default_operator(param_type)

  def is_valid_operator(self, param_type, operator_name):
    if not hasattr(param_type, operator_name):
      log.error("%s: %s is not a valid operator for Parameter type %s",
                self.name, operator_name, param_type.__name__)
      return False

    if operator_name[:3] not in ['op1','op2','op3','op4','opn']:
      log.error("%s: %s is not a valid operator for Parameter type %s",
                self.name, operator_name, param_type.__name__)
      return False

    num_parents_required = operator_name[2]
    if num_parents_required == 'n':
      return True

    num_parents_required = int(num_parents_required)
    minimum_number_of_parents = self.minimum_number_of_parents()

    if num_parents_required > minimum_number_of_parents:
      log.error("%s: %s for Parameter type %s requires more input configs "
                + "than minimum number of parents, %d, produced by this technique",
                self.name, operator_name, param_type.__name__, minimum_number_of_parents)
      return False

    return True

  @classmethod
  def generate_technique(cls, manipulator=None, *args, **kwargs):
    """
    generate a composable technique with random operators
    """
    from .manipulator import composable_operators
    # randomly select a composable technique to generate
    t = cls(*args, **kwargs)
    if manipulator is None:
      return t

    paramset = set()
    for p in manipulator.params:
      paramset.add(type(p))

    # add some random operator for each param
    operator_map = {}
    for param in paramset:
      operators = composable_operators(param, t.minimum_number_of_parents())
      # TODO - sometimes use "default" operator (don't choose an operator?
      # TODO - lower chance of picking op1_nop?
      ComposableEvolutionaryTechnique.add_to_map(operator_map, param, random.choice(operators))

    t.set_operator_map(operator_map)
    t.use_default_generated_name()
    return t


class RandomThreeParentsComposableTechnique(ComposableEvolutionaryTechnique):
  """
  based on DifferentialEvolution
  """

  def __init__(self, cr = 0.9, must_mutate_count=1, information_sharing=1, *pargs, **kwargs):
    super(RandomThreeParentsComposableTechnique, self).__init__(*pargs, **kwargs)
    self.cr = cr
    self.must_mutate_count = must_mutate_count
    self.information_sharing = information_sharing

  @classmethod
  def get_hyper_parameters(cls):
    return ['population_size', 'cr', 'must_mutate_count', 'information_sharing']

  def minimum_number_of_parents(self):
    return 4

  def get_parents(self, population):
    self.use_f = random.random()
    population.sort(key=_.timestamp) # sort population by timestamp

    # copy oldest
    cfg = self.manipulator.copy(population[0].config)

    shuffled_population = list(map(_.config, population[1:]))
    # mix in the global best configuration
    shuffled_population += ([self.get_global_best_configuration()]
                            * self.information_sharing)
    random.shuffle(shuffled_population)

    # return oldest configuration +_3 other configurations
    return [cfg] + shuffled_population[0:3]

  def update_population(self, config, population):
    # replace the oldest configuration if the new one is better.
    population.sort(key=_.timestamp)
    if self.lt(config, population[0].config):
      population[0].config = config

    # mark that oldest configuration is updated
    population[0].touch()

    return population

  def select_parameters(self, params):
    """
    randomly select a subset of parameters to operate on
    """
    random.shuffle(params)
    ret_list = params[:self.must_mutate_count]
    for param in params[self.must_mutate_count:]:
      if random.random() < self.cr:
        ret_list.append(param)
    return ret_list

  def get_default_operator(self, param_type):
    return {'op_name': 'op4_set_linear', 'args': [1.0, self.use_f, -self.use_f], 'kwargs': {}}

class GreedyComposableTechnique(ComposableEvolutionaryTechnique):
  """
  Always mixes in global best as parents
  """
  def __init__(self,
               mutation_rate = 0.1,
               must_mutate_count = 1,
               population_size = 10,
               *pargs, **kwargs):
    super(GreedyComposableTechnique, self).__init__(*pargs, **kwargs)
    self.mutation_rate = mutation_rate
    self.must_mutate_count = must_mutate_count
    self.population_size = population_size


  @classmethod
  def get_hyper_parameters(cls):
    return ['mutation_rate', 'must_mutate_count']

  def minimum_number_of_parents(self):
    # specify that we will return at least 4 cfgs from get_parents
    # this maximizes # of operators we can use
    return 4

  def get_parents(self, population):
    population.sort(key=_.timestamp) # sort population by timestamp

    # get a 50-50 mix of base and best cfgs as many operators do nothing given identical input cfgs
    cfg = self.manipulator.copy(population[0].config)
    # duplicate to get a total of 4 configurations to fulfill the promise in minimum_number_of_parents
    cfgs = [self.get_global_best_configuration(), cfg]*2
    # return a random 50-50 mix of the current configuration and global best to pass into operators
    random.shuffle(cfgs)
    return cfgs

  def update_population(self, config, population):
    # replace the oldest configuration if the new one is better.
    population.sort(key=_.timestamp)
    if self.lt(config, population[0].config):
      population[0].config = config

    # mark that oldest configuration is updated
    population[0].touch()

    return population

  def select_parameters(self, params):
    random.shuffle(params)
    ret_list = params[:self.must_mutate_count]
    for param in params[self.must_mutate_count:]:
      if random.random() < self.mutation_rate:
        ret_list.append(param)
    return ret_list

  def get_default_operator(self, param_type):
    return {'op_name': 'op1_randomize', 'args': [], 'kwargs':{}}


register(RandomThreeParentsComposableTechnique(name='ComposableDiffEvolution',
                                                 population_size=30))
register_generator(RandomThreeParentsComposableTechnique)
register_generator(GreedyComposableTechnique)


op_map = {}
ComposableEvolutionaryTechnique.add_to_map(op_map,
                                      PermutationParameter,
                                      "op3_cross", xchoice='op3_cross_CX')
ComposableEvolutionaryTechnique.add_to_map(op_map,
                                      "FloatArray",
                                      "op3_cross", strength=0.4)
register(RandomThreeParentsComposableTechnique(name='ComposableDiffEvolutionCX',
                                                 operator_map=op_map,
                                                 population_size=30))



from __future__ import division
from __future__ import absolute_import
from builtins import str
from builtins import range
from builtins import object
from past.utils import old_div
import abc
import copy
import logging
import math
import random
from collections import deque

from .metatechniques import MetaSearchTechnique
from .technique import register, SearchTechnique, all_techniques, get_random_generator_technique

log = logging.getLogger(__name__)


class BanditQueue(object):
  def __init__(self, keys, C=0.05, window=500, **kwargs):
    """
    C is exploration/exploitation tradeoff
    window is how long to remember past results
    """
    super(BanditQueue, self).__init__(**kwargs)
    self.C = C
    self.history = deque()
    self.keys = keys
    self.use_counts = dict(((k, 0) for k in keys))
    self.window = window
    self.request_count = 0

  @abc.abstractmethod
  def exploitation_term(self, key):
    """
    value 0 to 1.0 to represent quality of technique
    """
    return 0.0

  def exploration_term(self, key):
    """
    value represent how unsure we are (optimal bandit solution)
    """
    if self.use_counts[key] > 0:
      return math.sqrt(old_div((2.0 * math.log(len(self.history), 2.0)), self.use_counts[key]))
    else:
      return float('inf')

  def bandit_score(self, key):
    return (self.exploitation_term(key) +
            self.C * self.exploration_term(key))

  def ordered_keys(self):
    """select the next technique to use"""

    keys = list(self.keys)
    random.shuffle(keys)  # break ties randomly
    keys.sort(key=self.bandit_score)

    self.request_count += 1
    if log.isEnabledFor(logging.DEBUG) and (self.request_count % 1000) == 0:
      log.debug(str([
          (t, self.exploitation_term(t), self.C * self.exploration_term(t))
          for t in keys]))

    return reversed(keys)

  def on_result(self, key, value):
    self.history.append((key, value))
    self.on_push_history(key, value)
    if len(self.history) > self.window:
      self.on_pop_history(*self.history.popleft())

  def on_push_history(self, key, value):
    self.use_counts[key] += 1

  def on_pop_history(self, key, value):
    self.use_counts[key] -= 1


class AUCBanditQueue(BanditQueue):
  """
  Area Under the Receiving Operator Curve (AUC) credit assignment

  See:
  Comparison-based adaptive strategy selection with bandits in differential
  evolution. Fialho et al.
  """

  def __init__(self, *args, **kwargs):
    super(AUCBanditQueue, self).__init__(*args, **kwargs)
    self.debug = kwargs.get('debug', False)
    self.auc_sum = dict(((t, 0) for t in self.keys))
    self.auc_decay = dict(((t, 0) for t in self.keys))

  def exploitation_term_slow(self, key):
    """
    value 0 to 1.0 to represent quality of key

    computes the area under the curve where finding a new
    global best results in adding 1 to a cumulative total
    """
    score = 0.0
    pos = 0
    for t, value in self.history:
      if t is key:
        pos += 1
        if value:
          score += pos
    if pos:
      return score * 2.0 / (pos * (pos + 1.0))
    else:
      return 0.0

  def exploitation_term_fast(self, key):
    """
    value 0 to 1.0 to represent quality of key

    optimized O(1) implementation exploitation_term_slow()
    """
    score = self.auc_sum[key]
    pos = self.use_counts[key]
    if pos:
      return score * 2.0 / (pos * (pos + 1.0))
    else:
      return 0.0

  def exploitation_term(self, key):
    v1 = self.exploitation_term_fast(key)
    if self.debug:
      v2 = self.exploitation_term_slow(key)
      assert v1 == v2
    return v1

  def on_push_history(self, key, value):
    super(AUCBanditQueue, self).on_push_history(key, value)
    if value:
      self.auc_sum[key] += self.use_counts[key]
      self.auc_decay[key] += 1

  def on_pop_history(self, key, value):
    super(AUCBanditQueue, self).on_pop_history(key, value)
    self.auc_sum[key] -= self.auc_decay[key]
    if value:
      self.auc_decay[key] -= 1



class AUCBanditMetaTechnique(MetaSearchTechnique):
  def __init__(self, techniques, bandit_kwargs=dict(), **kwargs):
    super(AUCBanditMetaTechnique, self).__init__(techniques, **kwargs)
    self.bandit = AUCBanditQueue([t.name for t in techniques], **bandit_kwargs)
    self.name_to_technique = dict(((t.name, t) for t in self.techniques))

  def select_technique_order(self):
    """select the next technique to use"""
    return (self.name_to_technique[k] for k in self.bandit.ordered_keys())

  def on_technique_result(self, technique, result):
    self.bandit.on_result(technique.name, result.was_new_best)

  def on_technique_no_desired_result(self, technique):
    """treat not providing a configuration as not a best"""
    self.bandit.on_result(technique.name, 0)

  @classmethod
  def generate_technique(cls, manipulator=None, num_techniques=5, retry_count=3, generator_weight=10, *args, **kwargs):
    """
    Generate a bandit by randomly selecting existing techniques or composable techniques.
    If a composable technique is selected, the operators are then chosen

    :param manipulator: a ConfigurationManipulator used to enumerate parameters
    :param num_techniques: max number of subtechniques in the bandit
    :param retry_count: number of times to try getting a new technique before giving up
    :param generator_weight: weight to increase probability of choosing to generate a technique
    """
    techniques, generators = all_techniques()

    # get set of parameters to consider
    paramset = set()
    for p in manipulator.params:
      paramset.add(type(p))

    # filter techniques to get rid of metatechniques
    basetechniques = [t for t in techniques if not isinstance(t, MetaSearchTechnique)]
    bandit_techniques = []
    for i in range(num_techniques):
      for j in range(retry_count):
        # pick a technique or generate a composable
        if random.random() < old_div(float(len(basetechniques)), (len(basetechniques) + generator_weight*len(generators))):
          candidate = copy.deepcopy(random.choice(basetechniques))
        else:
          # pick a random generator
          candidate = get_random_generator_technique(generators, manipulator=manipulator)
        if not (candidate.name in [t.name for t in bandit_techniques]):
          bandit_techniques.append(candidate)
          break

    # make a bandit of the output list
    return cls(bandit_techniques, name="GeneratedBandit", *args, **kwargs)


class AUCBanditMutationTechnique(SearchTechnique):
  def __init__(self, bandit_kwargs=dict(), **kwargs):
    super(AUCBanditMutationTechnique, self).__init__(**kwargs)
    self.bandit = None
    self.bandit_kwargs = bandit_kwargs
    self.pending_results = []

  def handle_requested_result(self, result):
    for i in range(len(self.pending_results)):
      cfg, name, index = self.pending_results[i]
      if result.configuration == cfg:
        self.bandit.on_result((name, index), result.was_new_best)
        del self.pending_results[i]
        return
    log.warning("unexpected result")

  def desired_configuration(self):
    """
    use bandit to pick a single manipulator and apply it
    """
    seed = self.get_seed()
    if self.bandit is None:
      self.init_bandit(seed)

    cfg = self.manipulator.copy(seed)
    hash1 = self.manipulator.hash_config(cfg)
    params = self.manipulator.parameters_dict(cfg)
    for name, index in self.bandit.ordered_keys():
      if name in params:
        param = params[name]
        fns = param.manipulators(cfg)
        fn = fns[index % len(fns)]
        fn(cfg)
        hash2 = self.manipulator.hash_config(cfg)
        if hash1 != hash2:
          cfg = self.driver.get_configuration(cfg)
          self.pending_results.append((cfg, name, index))
          log.debug("applied %s[%s] manipulator function", name, index)
          return cfg

    return None


  def init_bandit(self, cfg):
    options = []
    for param in self.manipulator.parameters(cfg):
      for i in range(len(param.manipulators(cfg))):
        options.append((param.name, i))
    # TODO(jansel): remove assumption that set of parameters are fixed
    self.bandit = AUCBanditQueue(options, **self.bandit_kwargs)

  def get_seed(self):
    """seed mutation with global best"""
    if (self.driver.best_result is not None and
        self.driver.best_result.state == 'OK'):
      return self.driver.best_result.configuration.data
    else:
      return self.manipulator.random()


from . import evolutionarytechniques
from . import differentialevolution
from . import simplextechniques
from . import patternsearch
from . import simulatedannealing
from .pso import PSO, HybridParticle
from . import globalGA
register(AUCBanditMutationTechnique())

register(AUCBanditMetaTechnique([
        differentialevolution.DifferentialEvolutionAlt(),
        evolutionarytechniques.UniformGreedyMutation(),
        evolutionarytechniques.NormalGreedyMutation(mutation_rate=0.3),
        simplextechniques.RandomNelderMead(),
      ], name = "AUCBanditMetaTechniqueA"))
register(AUCBanditMetaTechnique([
        differentialevolution.DifferentialEvolutionAlt(),
        evolutionarytechniques.UniformGreedyMutation(),
      ], name = "AUCBanditMetaTechniqueB"))
register(AUCBanditMetaTechnique([
        differentialevolution.DifferentialEvolutionAlt(),
        patternsearch.PatternSearch(),
      ], name = "AUCBanditMetaTechniqueC"))
register(AUCBanditMetaTechnique([
        PSO(crossover = 'op3_cross_OX3'),
        PSO(crossover = 'op3_cross_OX1'),
        PSO(crossover = 'op3_cross_CX'),
        PSO(crossover = 'op3_cross_PMX'),
        PSO(crossover = 'op3_cross_PX'),
        evolutionarytechniques.GA(crossover = 'op3_cross_OX3', mutation_rate=0.01, crossover_rate=0.8),
        evolutionarytechniques.GA(crossover = 'op3_cross_OX1', mutation_rate=0.01, crossover_rate=0.8),
        evolutionarytechniques.GA(crossover = 'op3_cross_CX', mutation_rate=0.01, crossover_rate=0.8),
        evolutionarytechniques.GA(crossover = 'op3_cross_PX', mutation_rate=0.01, crossover_rate=0.8),
        evolutionarytechniques.GA(crossover = 'op3_cross_PMX', mutation_rate=0.01, crossover_rate=0.8),
        evolutionarytechniques.UniformGreedyMutation(name='ga-base', mutation_rate=0.01)
      ], name = "PSO_GA_Bandit"))
register(AUCBanditMetaTechnique([
	differentialevolution.DifferentialEvolutionAlt(),
	simulatedannealing.PseudoAnnealingSearch()
      ], name = "test"))
register(AUCBanditMetaTechnique([
        differentialevolution.DifferentialEvolutionAlt(),
        evolutionarytechniques.UniformGreedyMutation(),
        evolutionarytechniques.NormalGreedyMutation(mutation_rate=0.3),
        simplextechniques.RandomNelderMead(),
	simulatedannealing.PseudoAnnealingSearch()
      ], name = "test2"))
register(AUCBanditMetaTechnique([
	PSO(crossover='op3_cross_OX1'),
	PSO(crossover='op3_cross_PMX'),
	PSO(crossover='op3_cross_PX'),
	evolutionarytechniques.GA(crossover='op3_cross_OX1', crossover_rate=0.5),
	evolutionarytechniques.GA(crossover='op3_cross_PMX', crossover_rate=0.5),
	evolutionarytechniques.GA(crossover='op3_cross_PX', crossover_rate=0.5),
	differentialevolution.DifferentialEvolutionAlt(),
        globalGA.NormalGreedyMutation( crossover_rate=0.5, crossover_strength=0.2, name='GGA')
	], name='PSO_GA_DE'))




from __future__ import absolute_import

from . import measurement
from . import resultsdb
from . import search
from . import tuningrunmain
from opentuner.measurement import MeasurementInterface
from opentuner.resultsdb.models import Configuration
from opentuner.resultsdb.models import DesiredResult
from opentuner.resultsdb.models import Result
from opentuner.resultsdb.models import TuningRun
from opentuner.search.manipulator import ConfigurationManipulator
from opentuner.search.manipulator import EnumParameter
from opentuner.search.manipulator import FloatParameter
from opentuner.search.manipulator import IntegerParameter
from opentuner.search.manipulator import LogFloatParameter
from opentuner.search.manipulator import LogIntegerParameter
from opentuner.search.manipulator import PermutationParameter
from opentuner.search.manipulator import ScheduleParameter
from opentuner.search.manipulator import SwitchParameter
from opentuner.tuningrunmain import init_logging


def argparsers():
  """
  return a list of ArguementParser to be used as parents to the user's
  """
  return [
      measurement.driver.argparser,
      measurement.interface.argparser,
      search.driver.argparser,
      search.plugin.argparser,
      search.technique.argparser,
      #stats.argparser,
      tuningrunmain.argparser,
    ]


def default_argparser():
  import argparse
  return argparse.ArgumentParser(parents=argparsers())




from __future__ import print_function
from builtins import map
from builtins import object
import abc
import argparse
import logging
import time

from datetime import datetime
from fn import _
from future.utils import with_metaclass

log = logging.getLogger(__name__)
display_log = logging.getLogger(__name__ + ".DisplayPlugin")

argparser = argparse.ArgumentParser(add_help=False)
argparser.add_argument('--results-log',
    help="file to store log of the best configuration times")
argparser.add_argument('--results-log-details',
    help="file to store log of the non-best configuration times")
argparser.add_argument('--quiet', action='store_true',
    help="print less information")
argparser.add_argument('--display-frequency', default=10, type=int,
    help="how often for DisplayPlugin to print")

class SearchPlugin(object):
  @property
  def priority(self):
    """control order the plugin hooks gets run in, lower runs first"""
    return 0

  def set_driver(self, driver):
    """called before all other methods"""
    self.driver = driver

  def before_main(self): pass
  def after_main(self):  pass

  def before_techniques(self): pass
  def after_techniques(self):  pass

  def before_results_wait(self): pass
  def after_results_wait(self):  pass

  def on_result(self, result):
    """
    called once for every new result
    """
    pass

  def on_result_for_technique(self, result, technique):
    """
    called right before a result is given to a technique
    (result may be requested by multiple techniques)
    """
    pass

  def on_new_best_result(self, result):
    """
    called whenever the global best result changes
    """
    pass

class DisplayPlugin(with_metaclass(abc.ABCMeta, SearchPlugin)):
  def __init__(self, display_period=5):
    super(DisplayPlugin, self).__init__()
    self.last  = time.time()
    self.start = time.time()
    self.display_period = display_period

  def after_results_wait(self):
    t = time.time()
    if t - self.display_period > self.last:
      # call display every 5 seconds
      self.last = t
      self.display(t)

  def after_main(self):
    self.display()

  @abc.abstractmethod
  def display(self, t=None):
    pass


class LogDisplayPlugin(DisplayPlugin):
  def display(self, t=None):
    if not t:
      t = time.time()
    count = self.driver.results_query().count()
    best = self.driver.results_query(objective_ordered = True).first()
    if best is None:
      log.warning("no results yet")
      return
    requestor = ','.join(map(_.requestor, best.desired_results))
    display_log.info("tests=%d, best %s, cost %s, found by %s",
                     count,
                     cfg_repr(best.configuration),
                     self.driver.objective.display(best),
                     requestor,
                     )

class FileDisplayPlugin(SearchPlugin):
  def __init__(self, out, details, *args, **kwargs):
    super(FileDisplayPlugin, self).__init__(*args, **kwargs)
    self.last_best = float('inf')
    self.start_date = datetime.now()
    if out:
      self.out = open(out, "w")
    else:
      self.out = None
    if out == details:
      self.details = self.out
      self.out = None
    elif details:
      self.details = open(details, "w")
    else:
      self.details = None

  def on_result(self, result):
    if self.out and result.time < self.last_best:
      self.last_best = result.time
      print((result.collection_date - self.start_date).total_seconds(), \
          result.time, file=self.out)
      self.out.flush()
    if self.details:
      print((result.collection_date - self.start_date).total_seconds(), \
          result.time, file=self.details)
      self.details.flush()

def get_enabled(args):
  plugins = []
  if not args.quiet:
    plugins.append(LogDisplayPlugin(args.display_frequency))
  if args.results_log or args.results_log_details:
    plugins.append(FileDisplayPlugin(args.results_log,
                                     args.results_log_details))
  return plugins

def cfg_repr(cfg):
  try:
    s = repr(cfg.data)
    if len(s) < 100:
      return s
  except:
    pass
  return "#{0}".format(cfg.id)










from past.builtins import execfile
import sys
from os.path import normpath, realpath, dirname, join, isfile

project_root = normpath(join(dirname(realpath(__file__)), '../..'))

if 'venv' not in ','.join(sys.path):
  venv_activate = join(project_root, 'venv/bin/activate_this.py')
  if isfile(venv_activate):
    execfile(venv_activate, dict(__file__=venv_activate))

sys.path.insert(0, project_root)




from __future__ import division
from builtins import range
from past.utils import old_div
from opentuner.search import technique
import math
import random
#Default interval steps for cooling schedules
DEFAULT_INTERVAL = 100

#Pseudo-annealing - no relative energy input into acceptance function
class PseudoAnnealingSearch(technique.SequentialSearchTechnique):
  def __init__(self,
               temps = [30,0], #temperature schedule
               intervals = [],  #duration schedule
          		 loop = True, #Do we loop the schedule if we reach the end?
               *pargs, **kwargs):
    #fill intervals sufficiently
    ext_intervals = list(intervals)
    for i in range(len(temps)-len(intervals)-1):
      ext_intervals.append(DEFAULT_INTERVAL)
            
    #create temperature schedule (list of temps)
    cool_schedule = [temps[0]]
    for i in range(len(temps)-1):
      step = old_div((float(temps[i+1]) - temps[i]),ext_intervals[i])
      for j in range(ext_intervals[i]):
        cool_schedule.append(max(cool_schedule[-1] + step,0))
      
    self.cool_schedule = cool_schedule
    self.loop = loop
    self.scaling = 50 #scaling of acceptance function
      
    super(PseudoAnnealingSearch,self).__init__(*pargs,**kwargs)


  def main_generator(self):
    objective = self.objective
    driver = self.driver
    manipulator = self.manipulator

    #Start in a random spot
    state = driver.get_configuration(manipulator.random())
    yield state
    #schedule counter
    counter = 0
    max_time = len(self.cool_schedule)-1
    #Check whether relative objective implemented
    has_rel = objective.relative(state,state) is not None
    has_rel=False
              
    while True:
      #Determine temperature
      temp = self.cool_schedule[min(counter,max_time)]
      #scale stepsize with temp and time (arbitrary)
      step_size = math.exp(old_div(-(20 + old_div(counter,100)),(temp+ 1))) 
          
      #get candidate neighbors using manipulator
      points = list()
      points.append(state)
      for param in manipulator.parameters(state.data):
        if param.is_primitive():
          # get current value of param, scaled to be in range [0.0, 1.0]
          unit_value = param.get_unit_value(state.data)
          if unit_value > 0.0:
            # produce new config with param set step_size lower
            down_cfg = manipulator.copy(state.data)
            param.set_unit_value(down_cfg, max(0.0, unit_value - step_size*random.random()))
            down_cfg = driver.get_configuration(down_cfg)
            self.yield_nonblocking(down_cfg)
            points.append(down_cfg)

          if unit_value < 1.0:
            # produce new config with param set step_size higher
            up_cfg = manipulator.copy(state.data)
            param.set_unit_value(up_cfg, min(1.0, unit_value + step_size*random.random()))
            up_cfg = driver.get_configuration(up_cfg)
            self.yield_nonblocking(up_cfg)
            points.append(up_cfg)
        else: # ComplexParameter
          for mutate_function in param.manipulators(state.data):
            cfg = manipulator.copy(state.data)
            mutate_function(cfg)
            cfg = driver.get_configuration(cfg)
            self.yield_nonblocking(cfg)
            points.append(cfg)
      yield None # wait for all results
            
      #Relative comparison implemented
      if has_rel:
        while True:
          if len(points) == 0:
            state = driver.best_result.configuration
            break
          candidate = points.pop(random.randint(0,len(points)-1))
          #compare to global best
          if random.random() < AcceptanceFunction(1, objective.relative(candidate,driver.best_result.configuration), temp, self.scaling):
            state = candidate
            break
      #No relative compare
      else:
      #sort points by "energy" (quality)
        points.sort(cmp=objective.compare)
            
        #Make decision about changing state
        #probability picking next-best state is exp^(-1/temp)
        #repeat and cycle to get state p-dist resembling this
        sel = 0
        while AcceptanceFunction(0,1,temp,1)>random.random():
          sel += 1
        state = points[sel%len(points)]
            
        #switch to the global best if temperature is low (i.e. we aren't moving much)
        if AcceptanceFunction(0,1,temp,1)< .0001 and objective.lt(driver.best_result.configuration, state):
          state = driver.best_result.configuration
          
      #update counter
      counter +=1
      if counter>max_time and self.loop:
        counter=counter-max_time
              

#Acceptance probability function for annealing
def AcceptanceFunction(e,e_new,temp,scaling):
  #Standard acceptance probability function using relative "goodness"
  if e>=e_new:
    return 1
  if temp == 0:
    return 0
  if scaling*(e_new-e)/temp > 10:
    #for practical purposes, probability is too low.
    return 0
  return math.exp(scaling*(e-e_new)/temp)


#register technique
technique.register(PseudoAnnealingSearch())



from builtins import range
from builtins import object
import abc
import argparse
import errno
import hashlib
import logging
import os
import re
import signal
import subprocess
import threading
import time
from multiprocessing.pool import ThreadPool
from future.utils import with_metaclass

try:
  import resource
except ImportError:
  resource = None

try:
  import fcntl
except ImportError:
  fcntl = None

import opentuner
from opentuner import resultsdb
from opentuner.resultsdb.models import *

log = logging.getLogger(__name__)

argparser = argparse.ArgumentParser(add_help=False)
argparser.add_argument('--parallel-compile', action='store_true',
                       default=False,
                       help="present if compiling can be done in parallel")

the_io_thread_pool = None


class MeasurementInterface(with_metaclass(abc.ABCMeta, object)):
  """
  abstract base class for compile and measurement
  """

  def __init__(self,
               args=None,
               project_name=None,
               program_name='unknown',
               program_version='unknown',
               manipulator=None,
               objective=None,
               input_manager=None):
    self.args = args
    self._project = project_name
    self._program = program_name
    self._version = program_version
    self._objective = objective
    self._manipulator = manipulator
    self._input_manager = input_manager

    self.pids = []
    self.pid_lock = threading.Lock()
    self.parallel_compile = args.parallel_compile
    # If parallel_compile is False then compile_and_run() will be invoked
    # sequentially otherwise the driver first invokes compile() in parallel
    # followed by run_precompiled() sequentially

  def compile(self, config_data, id):
    """
    Compile in PARALLEL according to the configuration in config_data 
    (obtained from desired_result.configuration) Should use id parameter 
    to determine output location of executable Return value will be passed 
    to run_precompiled as compile_result, useful for storing error/timeout 
    information
    """
    if self.parallel_compile:
        raise RuntimeError('MeasurementInterface.compile() not implemented for',
                'parallel compilation')
    pass

  def run_precompiled(self, desired_result, input, limit, compile_result, id):
    """
    Run the given desired_result SEQUENTIALLY on input and produce a Result() 
    Abort early if limit (in seconds) is reached Assume that the executable
    to be measured has already been compiled to an executable corresponding to
    identifier id by compile() The compile_result is the return result of compile(), 
    and it will be None if compile() was not called
    """
    if self.parallel_compile:
        raise RuntimeError('MeasurementInterface.run_precompiled() not implemented', 
                'for parallel compilation')
    pass

  def cleanup(self, id):
    """
    Clean up any temporary files associated with the executable
    """
    pass

  def pre_process(self):
    """
    The process before each iteration This method will be called
    once per iteration before all threads are launched
    """
    pass

  def post_process(self):
    """
    The process after each iteration This method will be called 
    once per iteration after all threads are committed
    """
    pass

  def extra_convergence_criteria(self, result):
    """
    The extra convergence criteria which returns True if the
    current result is acceptable by the user
    """
    return False

  #@abc.abstractmethod
  def compile_and_run(self, desired_result, input, limit):
    """
    Compile and run the given desired_result on input and produce a 
    Result(), abort early if limit (in seconds) is reached This function 
    is only used for sequential execution flow

    FIXME: Shoud uncomment @abc.abstractmethod Now comment out for
    compatiability
    """
    return self.run(desired_result, input, limit)

  def run(self, desired_result, input, limit):
    """
    run the given desired_result on input and produce a Result(),
    abort early if limit (in seconds) is reached
    """
    return opentuner.resultdb.models.Result()

  def save_final_config(self, config):
    """
    called at the end of autotuning with the best resultsdb.models.Configuration
    """
    try:
      config_str = repr(config.data)
      if len(config_str) > 256:
        config_str = config_str[:256] + '...'
      log.info('final configuration: %s', config_str)
      log.info('you may want to implement save_final_config(), to store this')
    except:
      log.error('error printing configuration', exc_info=True)

  def db_program_version(self, session):
    """return a version identifier for the program being tuned"""
    return resultsdb.models.ProgramVersion.get(
        session=session,
        project=self.project_name(),
        name=self.program_name(),
        version=self.program_version(),
        parameter_info=self.manipulator().parameters_to_json(),
    )

  def set_driver(self, measurement_driver):
    self.driver = measurement_driver

  def project_name(self):
    if self._project is not None:
      return self._project
    autoname = re.sub('(Measurement?)Interface$', '', self.__class__.__name__)
    if autoname:
      return autoname
    else:
      return 'unknown'

  def program_name(self):
    return self._program

  def program_version(self):
    return self._version

  def file_hash(self, filename):
    """helper used to generate program versions"""
    return hashlib.sha256(open(filename).read()).hexdigest()

  def manipulator(self):
    """
    called once to create the search.manipulator.ConfigurationManipulator
    """
    if self._manipulator is None:
      msg = ('MeasurementInterface.manipulator() must be implemented or a '
             '"manipulator=..." must be provided to the constructor')
      log.error(msg)
      raise Exception(msg)
    return self._manipulator

  def objective(self):
    """
    called once to create the search.objective.SearchObjective
    """
    if self._objective is None:
      from ..search.objective import MinimizeTime

      return MinimizeTime()
    return self._objective

  def input_manager(self):
    """
    called once to create the measurement.inputmanager.InputManager
    """
    if self._objective is None:
      from .inputmanager import FixedInputManager

      return FixedInputManager()
    return self._input_manager

  def seed_configurations(self):
    """
    Extra seed configuration objects to add to those given on the command line.
    Configuration objects (typically dictionaries) not database objects.
    """
    return []

  def kill_all(self):
    self.pid_lock.acquire()
    for pid in self.pids:
      goodkillpg(pid)
    self.pids = []
    self.pid_lock.release()

  def call_program(self, cmd, limit=None, memory_limit=None, **kwargs):
    """
    call cmd and kill it if it runs for longer than limit

    returns dictionary like
      {'returncode': 0,
       'stdout': '', 'stderr': '',
       'timeout': False, 'time': 1.89}
    """
    the_io_thread_pool_init(self.args.parallelism)
    if limit is float('inf'):
      limit = None
    if type(cmd) in (str, str):
      kwargs['shell'] = True
    killed = False
    t0 = time.time()
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                         preexec_fn=preexec_setpgid_setrlimit(memory_limit),
                         **kwargs)
    # Add p.pid to list of processes to kill in case of keyboardinterrupt
    self.pid_lock.acquire()
    self.pids.append(p.pid)
    self.pid_lock.release()

    try:
      stdout_result = the_io_thread_pool.apply_async(p.stdout.read)
      stderr_result = the_io_thread_pool.apply_async(p.stderr.read)
      while p.returncode is None:
        if limit is None:
          goodwait(p)
        elif limit and time.time() > t0 + limit:
          killed = True
          goodkillpg(p.pid)
          goodwait(p)
        else:
          # still waiting...
          sleep_for = limit - (time.time() - t0)
          if not stdout_result.ready():
            stdout_result.wait(sleep_for)
          elif not stderr_result.ready():
            stderr_result.wait(sleep_for)
          else:
            #TODO(jansel): replace this with a portable waitpid
            time.sleep(0.001)
        p.poll()
    except:
      if p.returncode is None:
        goodkillpg(p.pid)
      raise
    finally:
      # No longer need to kill p
      self.pid_lock.acquire()
      if p.pid in self.pids:
        self.pids.remove(p.pid)
      self.pid_lock.release()

    t1 = time.time()
    return {'time': float('inf') if killed else (t1 - t0),
            'timeout': killed,
            'returncode': p.returncode,
            'stdout': stdout_result.get(),
            'stderr': stderr_result.get()}

  def prefix_hook(self, session):
    pass

  @classmethod
  def main(cls, args, *pargs, **kwargs):
    from opentuner.tuningrunmain import TuningRunMain

    return TuningRunMain(cls(args, *pargs, **kwargs), args).main()


class DefaultMeasurementInterface(MeasurementInterface):
  def run(self, desired_result, input, limit):
    raise RuntimeError('MeasurementInterface.run() not implemented')


def preexec_setpgid_setrlimit(memory_limit):
  if resource is not None:
    def _preexec():
      os.setpgid(0, 0)
      try:
        resource.setrlimit(resource.RLIMIT_CORE, (0, 0))
      except ValueError:
        pass  # No permission
      if memory_limit:
        try:
          (soft, hard) = resource.getrlimit(resource.RLIMIT_AS)
          resource.setrlimit(resource.RLIMIT_AS, (min(soft, memory_limit),
                                                  min(hard, memory_limit)))
        except ValueError:
          pass  # No permission
    return _preexec


def the_io_thread_pool_init(parallelism=1):
  global the_io_thread_pool
  if the_io_thread_pool is None:
    the_io_thread_pool = ThreadPool(2 * parallelism)
    # make sure the threads are started up
    the_io_thread_pool.map(int, list(range(2 * parallelism)))


def goodkillpg(pid):
  """
  wrapper around kill to catch errors
  """
  log.debug("killing pid %d", pid)
  try:
    if hasattr(os, 'killpg'):
      os.killpg(pid, signal.SIGKILL)
    else:
      os.kill(pid, signal.SIGKILL)
  except:
    log.error('error killing process %s', pid, exc_info=True)


def goodwait(p):
  """
  python doesn't check if its system calls return EINTR, retry if it does
  """
  while True:
    try:
      rv = p.wait()
      return rv
    except OSError as e:
      if e.errno != errno.EINTR:
        raise




from future import standard_library
standard_library.install_aliases()
from builtins import object
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy import create_engine
from sqlalchemy.orm import relationship, backref
from sqlalchemy import (
  Column, Integer, String, DateTime, Boolean, Enum,
  Float, PickleType, ForeignKey, Text, func, Index)
import sqlalchemy
import re

from pickle import dumps, loads
from gzip import zlib
class CompressedPickler(object):
  @classmethod
  def dumps(cls, obj, protocol=2):
    s = dumps(obj, protocol)
    sz = zlib.compress(s, 9)
    if len(sz) < len(s):
      return sz
    else:
      return s

  @classmethod
  def loads(cls, string):
    try:
      s = zlib.decompress(string)
    except:
      s = string
    return loads(s)

class Base(object):
  @declared_attr
  def __tablename__(cls):
    """convert camel case to underscores"""
    return re.sub(r'([a-z])([A-Z])', r'\1_\2', cls.__name__).lower()

  id = Column(Integer, primary_key=True, index=True)


Base = declarative_base(cls=Base)

class _Meta(Base):
  """ meta table to track current version """
  db_version = Column(String(128))

  @classmethod
  def has_version(cls, session, version):
    try:
      session.flush()
      session.query(_Meta).filter_by(db_version=version).one()
      return True
    except sqlalchemy.orm.exc.NoResultFound:
      return False

  @classmethod
  def get_version(cls, session):
    try:
      session.flush()
      x = session.query(_Meta).one()
      return x.db_version
    except sqlalchemy.orm.exc.NoResultFound:
      return None

  @classmethod
  def add_version(cls, session, version):
    if not cls.has_version(session, version):
      session.add(_Meta(db_version=version))


class Program(Base):
  project = Column(String(128))
  name = Column(String(128))

  @classmethod
  def get(cls, session, project, name):
    try:
      session.flush()
      return session.query(Program).filter_by(project=project, name=name).one()
    except sqlalchemy.orm.exc.NoResultFound:
      t = Program(project=project, name=name)
      session.add(t)
      return t


class ProgramVersion(Base):
  program_id = Column(ForeignKey(Program.id))
  program = relationship(Program, backref='versions')
  version = Column(String(128))
  parameter_info = Column(Text)

  @property
  def name(self):
    return self.program.name

  @property
  def project(self):
    return self.program.project

  @classmethod
  def get(cls, session, project, name, version, parameter_info=None):
    program = Program.get(session, project, name)
    try:
      session.flush()
      if parameter_info is None:
        return session.query(ProgramVersion).filter_by(program=program,
                                                     version=version).one()
      else:
        return session.query(ProgramVersion).filter_by(program=program,
                                                      version=version,
                                                      parameter_info=parameter_info).one()
    except sqlalchemy.orm.exc.NoResultFound:
      t = ProgramVersion(program=program, version=version, parameter_info=parameter_info)
      session.add(t)
      return t


class Configuration(Base):
  program_id = Column(ForeignKey(Program.id))
  program = relationship(Program)
  hash = Column(String(64))
  data = Column(PickleType(pickler=CompressedPickler))

  @classmethod
  def get(cls, session, program, hashv, datav):
    try:
      session.flush()
      return (session.query(Configuration)
              .filter_by(program=program, hash=hashv).one())
    except sqlalchemy.orm.exc.NoResultFound:
      t = Configuration(program=program, hash=hashv, data=datav)
      session.add(t)
      return t


Index('ix_configuration_custom1', Configuration.program_id, Configuration.hash)


class MachineClass(Base):
  name = Column(String(128))

  @classmethod
  def get(cls, session, name):
    try:
      session.flush()
      return session.query(MachineClass).filter_by(name=name).one()
    except sqlalchemy.orm.exc.NoResultFound:
      t = MachineClass(name=name)
      session.add(t)
      return t


class Machine(Base):
  name = Column(String(128))

  cpu = Column(String(128))
  cores = Column(Integer)
  memory_gb = Column(Float)

  machine_class_id = Column(ForeignKey(MachineClass.id))
  machine_class = relationship(MachineClass, backref='machines')


class InputClass(Base):
  program_id = Column(ForeignKey(Program.id))
  program = relationship(Program, backref='inputs')

  name = Column(String(128))
  size = Column(Integer)

  @classmethod
  def get(cls, session, program, name='default', size=-1):
    try:
      session.flush()
      return session.query(InputClass).filter_by(program=program,
                                                 name=name,
                                                 size=size).one()
    except sqlalchemy.orm.exc.NoResultFound:
      t = InputClass(program=program, name=name, size=size)
      session.add(t)
      return t


class Input(Base):
  #state          = Column(Enum('ANY_MACHINE', 'SINGLE_MACHINE', 'DELETED'),
  #                        default='ANY_MACHINE', name='t_input_state')

  input_class_id = Column(ForeignKey(InputClass.id))
  input_class = relationship(InputClass, backref='inputs')

  #optional, set only for state='SINGLE_MACHINE'
  #machine_id     = Column(ForeignKey(MachineClass.id))
  #machine        = relationship(MachineClass, backref='inputs')

  #optional, for use by InputManager
  path = Column(Text)
  extra = Column(PickleType(pickler=CompressedPickler))


class TuningRun(Base):
  uuid = Column(String(32), index=True, unique=True)

  program_version_id = Column(ForeignKey(ProgramVersion.id))
  program_version = relationship(ProgramVersion, backref='tuning_runs')

  machine_class_id = Column(ForeignKey(MachineClass.id))
  machine_class = relationship(MachineClass, backref='tuning_runs')

  input_class_id = Column(ForeignKey(InputClass.id))
  input_class = relationship(InputClass, backref='tuning_runs')

  name = Column(String(128), default='unnamed')
  args = Column(PickleType(pickler=CompressedPickler))
  objective = Column(PickleType(pickler=CompressedPickler))

  state = Column(Enum('QUEUED', 'RUNNING', 'COMPLETE', 'ABORTED',
                      name='t_tr_state'),
                 default='QUEUED')
  start_date = Column(DateTime, default=func.now())
  end_date = Column(DateTime)

  final_config_id = Column(ForeignKey(Configuration.id))
  final_config = relationship(Configuration)

  #__mapper_args__ = {'primary_key': uuid}

  @property
  def program(self):
    return self.program_version.program


class Result(Base):
  #set by MeasurementDriver:
  configuration_id = Column(ForeignKey(Configuration.id))
  configuration = relationship(Configuration)

  machine_id = Column(ForeignKey(Machine.id))
  machine = relationship(Machine, backref='results')

  input_id = Column(ForeignKey(Input.id))
  input = relationship(Input, backref='results')

  tuning_run_id = Column(ForeignKey(TuningRun.id), index=True)
  tuning_run = relationship(TuningRun, backref='results')

  collection_date = Column(DateTime, default=func.now())
  collection_cost = Column(Float)

  #set by MeasurementInterface:
  state = Column(Enum('OK', 'TIMEOUT', 'ERROR',
                      name='t_result_state'),
                 default='OK')
  time = Column(Float)
  accuracy = Column(Float)
  energy = Column(Float)
  size = Column(Float)
  confidence = Column(Float)
  #extra = Column(PickleType)

  #set by SearchDriver
  was_new_best = Column(Boolean)


Index('ix_result_custom1', Result.tuning_run_id, Result.was_new_best)


class DesiredResult(Base):
  #set by the technique:
  configuration_id = Column(ForeignKey(Configuration.id))
  configuration = relationship(Configuration)
  limit = Column(Float)

  #set by the search driver
  priority = Column(Float)
  tuning_run_id = Column(ForeignKey(TuningRun.id))
  tuning_run = relationship(TuningRun, backref='desired_results')
  generation = Column(Integer)
  requestor = Column(String(128))
  request_date = Column(DateTime, default=func.now())

  #set by the measurement driver
  state = Column(Enum('UNKNOWN', 'REQUESTED', 'RUNNING',
                      'COMPLETE', 'ABORTED',
                      name="t_dr_state"),
                 default='UNKNOWN')
  result_id = Column(ForeignKey(Result.id), index=True)
  result = relationship(Result, backref='desired_results')
  start_date = Column(DateTime)

  #input_id        = Column(ForeignKey(Input.id))
  #input           = relationship(Input, backref='desired_results')


Index('ix_desired_result_custom1', DesiredResult.tuning_run_id,
      DesiredResult.generation)

Index('ix_desired_result_custom2', DesiredResult.tuning_run_id,
      DesiredResult.configuration_id)


# track bandit meta-technique information if a bandit meta-technique is used for a tuning run.
class BanditInfo(Base):
  tuning_run_id = Column(ForeignKey(TuningRun.id))
  tuning_run = relationship(TuningRun, backref='bandit_info')
  # the bandit exploration/exploitation tradeoff
  c = Column(Float)
  # the bandit window
  window = Column(Integer)

class BanditSubTechnique(Base):
  bandit_info_id = Column(ForeignKey(BanditInfo.id))
  bandit_info = relationship(BanditInfo, backref='subtechniques')
  name = Column(String(128))


if __name__ == '__main__':
  #test:
  engine = create_engine('sqlite:///:memory:', echo=True)
  Base.metadata.create_all(engine)




from datetime import datetime
from opentuner import tuningrunmain


class TuningRunManager(tuningrunmain.TuningRunMain):
  """
  This class manages a tuning run in a "slave" configuration, where main()
  is controlled by an another program.
  """
  def __init__(self, measurement_interface, args, **kwargs):
    super(TuningRunManager, self).__init__(measurement_interface, args, **kwargs)
    self.init()
    self.tuning_run.state = 'RUNNING'
    self.commit(force=True)
    self.search_driver.external_main_begin()

  def get_next_desired_result(self):
    """
    Returns a opentuner.resultsdb.DesiredResult that should be tested next.
    """
    dr = self.measurement_driver.query_pending_desired_results().first()
    if dr is None:
      self.search_driver.external_main_generation()
      dr = self.measurement_driver.query_pending_desired_results().first()
      if dr is None:
        return None
    self.measurement_driver.claim_desired_result(dr)
    dr.limit = self.measurement_driver.run_time_limit(dr)
    return dr

  def get_desired_results(self):
    """
    Returns a list of all opentuner.resultsdb.DesiredResult that should be tested next.
    """
    drs = self.measurement_driver.query_pending_desired_results().all()
    if len(drs) == 0:
      self.search_driver.external_main_generation()
      drs = self.measurement_driver.query_pending_desired_results().all()
      if len(drs) == 0:
        return []
    for dr in drs:
      self.measurement_driver.claim_desired_result(dr)
      dr.limit = self.measurement_driver.run_time_limit(dr)

    return drs

  def report_result(self, desired_result, result, result_input=None):
    """
    Report a measured result.  desired_result should have been returned by
    get_next_desired_result().
    """
    self.measurement_driver.report_result(desired_result, result, result_input)

  def get_best_configuration(self):
    """
    The best configuration found so far.  From the current tuning run only.
    """
    try:
      return self.search_driver.best_result.configuration.data
    except AttributeError:
      return None

  def get_best_result(self):
    """
    The best result found so far.  From the current tuning run only.
    """
    try:
      return self.search_driver.best_result
    except AttributeError:
      return None

  def finish(self):
    """
    Called at the end of the tuning process to call hooks and close database
    connections.
    """
    self.search_driver.external_main_end()
    self.measurement_interface.save_final_config(
        self.search_driver.best_result.configuration)
    self.tuning_run.final_config = self.search_driver.best_result.configuration
    self.tuning_run.state = 'COMPLETE'
    self.tuning_run.end_date = datetime.now()
    self.commit(force=True)
    self.session.close()








from opentuner.search import technique

class PatternSearch(technique.SequentialSearchTechnique):
  def main_generator(self):

    objective   = self.objective
    driver      = self.driver
    manipulator = self.manipulator

    # start at a random position
    center = driver.get_configuration(manipulator.random())
    self.yield_nonblocking(center)

    # initial step size is arbitrary
    step_size = 0.1

    while True:
      points = list()
      for param in manipulator.parameters(center.data):
        if param.is_primitive():
          # get current value of param, scaled to be in range [0.0, 1.0]
          unit_value = param.get_unit_value(center.data)

          if unit_value > 0.0:
            # produce new config with param set step_size lower
            down_cfg = manipulator.copy(center.data)
            param.set_unit_value(down_cfg, max(0.0, unit_value - step_size))
            down_cfg = driver.get_configuration(down_cfg)
            self.yield_nonblocking(down_cfg)
            points.append(down_cfg)

          if unit_value < 1.0:
            # produce new config with param set step_size higher
            up_cfg = manipulator.copy(center.data)
            param.set_unit_value(up_cfg, min(1.0, unit_value + step_size))
            up_cfg = driver.get_configuration(up_cfg)
            self.yield_nonblocking(up_cfg)
            points.append(up_cfg)

        else: # ComplexParameter
          for mutate_function in param.manipulators(center.data):
            cfg = manipulator.copy(center.data)
            mutate_function(cfg)
            cfg = driver.get_configuration(cfg)
            self.yield_nonblocking(cfg)
            points.append(cfg)


      yield None # wait for all results

      #sort points by quality, best point will be points[0], worst is points[-1]
      points.sort(cmp=objective.compare)

      if (objective.lt(driver.best_result.configuration, center)
          and driver.best_result.configuration != points[0]):
        # another technique found a new global best, switch to that
        center = driver.best_result.configuration
      elif objective.lt(points[0], center):
        # we found a better point, move there
        center = points[0]
      else:
        # no better point, shrink the pattern
        step_size /= 2.0

# register our new technique in global list
technique.register(PatternSearch())







#!/usr/bin/env python

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from builtins import str
from builtins import zip
from builtins import map
from builtins import range
from builtins import object
from past.utils import old_div
from functools import reduce
if __name__ == '__main__':
  from . import adddeps

import argparse
import csv
import hashlib
import itertools
import logging
import math
import os
import sqlalchemy.orm.exc
import subprocess
import sys

from collections import defaultdict
from fn import _
from fn import Stream
from fn.iters import repeat
from pprint import pprint

import opentuner
from opentuner import resultsdb
from opentuner.resultsdb.models import *

log = logging.getLogger('opentuner.utils.stats')

argparser = argparse.ArgumentParser()
argparser.add_argument('--label')
argparser.add_argument('--stats', action='store_true',
                       help="run in stats mode")
argparser.add_argument('--by-request-count', action='store_true',
                       help='report stats by request count')
argparser.add_argument('--stats-quanta', type=float, default=10,
                       help="step size in seconds for binning with --stats")
argparser.add_argument('--stats-dir', default='stats',
                       help="directory to output --stats to")
argparser.add_argument('--stats-input', default="opentuner.db")
argparser.add_argument('--min-runs',  type=int, default=1,
                       help="ignore series with less then N runs")

PCTSTEPS = list(map(old_div(_,20.0), list(range(21))))

def mean(vals):
  n = 0.0
  d = 0.0
  for v in vals:
    if v is not None:
      n += v
      d += 1.0
  if d == 0.0:
    return None
  return old_div(n,d)

def median(vals):
  vals = sorted(vals)
  a = old_div((len(vals)-1),2)
  b = old_div((len(vals)),2)
  return old_div((vals[a]+vals[b]),2.0)

def percentile(vals, pct):
  vals = sorted(vals)
  pos = (len(vals)-1) * pct
  a = int(math.floor(pos))
  b = min(len(vals) - 1, a + 1)
  return (1.0-(pos-a))*vals[a] + (pos-a)*vals[b]

def variance(vals):
  vals = [x for x in vals if x is not None]
  avg = mean(vals)
  if avg is None:
    return None
  if avg in (float('inf'), float('-inf')):
    return avg
  return mean(list(map((_ - avg) ** 2, vals)))

def stddev(vals):
  var = variance(vals)
  if var is None:
    return None
  return math.sqrt(var)

def hash_args(x):
  d = dict(vars(x))
  for k in ('database', 'results_log', 'results_log_details'):
    d[k] = None
  return hashlib.sha256(str(sorted(d.items()))).hexdigest()[:20]

def run_label(tr, short = False):
  techniques = ','.join(tr.args.technique)
  if not tr.name or tr.name=='unnamed':
    if short:
      return techniques
    else:
      return "%s_%s" % (techniques, hash_args(tr.args)[:6])
  else:
    return tr.name

def run_dir(base, tr):
  return os.path.join(base,
                      tr.program.project,
                      tr.program.name.split('/')[-1],
                      tr.program_version.version[:16])

class StatsMain(object):
  def __init__(self, args):
    self.args = args
    path = args.stats_input
    self.dbs = list()
    for f in os.listdir(path):
      if 'journal' in f:
        continue
      try:
        e, sm = resultsdb.connect('sqlite:///'+os.path.join(path, f))
        self.dbs.append(sm())
      except:
        log.error('failed to load database: %s', 
                  os.path.join(path, f),
                  exc_info=True)

  def main(self):
    dir_label_runs = defaultdict(lambda: defaultdict(list))
    for session in self.dbs:
      q = (session.query(resultsdb.models.TuningRun)
          .filter_by(state='COMPLETE')
          .order_by('name'))

      if self.args.label:
        q = q.filter(TuningRun.name.in_(
          list(map(str.strip,self.args.label.split(',')))))

      for tr in q:
        d = run_dir(self.args.stats_dir, tr)
        d = os.path.normpath(d)
        dir_label_runs[d][run_label(tr)].append((tr, session))

    summary_report = defaultdict(lambda: defaultdict(list))
    for d, label_runs in list(dir_label_runs.items()):
      if not os.path.isdir(d):
        os.makedirs(d)
      session = list(label_runs.values())[0][0][1]
      objective = list(label_runs.values())[0][0][0].objective
      all_run_ids = list(map(_[0].id, itertools.chain(*list(label_runs.values()))))
      q = (session.query(Result)
           .filter(Result.tuning_run_id.in_(all_run_ids))
           .filter(Result.time < float('inf'))
           .filter_by(was_new_best=True, state='OK'))
      total = q.count()
      if total == 0: 
          continue
      q = objective.filter_acceptable(q)
      acceptable = q.count()
      q = q.order_by(*objective.result_order_by_terms())
      best = q.limit(1).one()
      worst = q.offset(acceptable-1).limit(1).one()

      list(map(len, list(label_runs.values())))

      log.info("%s -- best %.4f / worst %.f4 "
               "-- %d of %d acceptable -- %d techniques with %d to %d runs",
               d,
               best.time,
               worst.time,
               acceptable,
               total,
               len(list(label_runs.values())),
               min(list(map(len, list(label_runs.values())))),
               max(list(map(len, list(label_runs.values())))))

      for label, runs in sorted(label_runs.items()):
        if len(runs) < self.args.min_runs:
          print(len(runs) ,self.args.min_runs)
          continue
        log.debug('%s/%s has %d runs %s',d, label, len(runs), runs[0][0].args.technique)
        self.combined_stats_over_time(d, label, runs, objective, worst, best)

        final_scores = list()
        for run, session in runs:
          try:
            final = (session.query(Result)
                    .filter_by(tuning_run=run,
                               configuration=run.final_config)
                    .limit(1)
                    .one())
          except sqlalchemy.orm.exc.NoResultFound:
            continue
          final_scores.append(objective.stats_quality_score(final, worst, best))
        final_scores.sort()
        if final_scores:
          norm = objective.stats_quality_score(best, worst, best)
          if norm > 0.00001:
            summary_report[d][run_label(run, short=True)] = (
                old_div(percentile(final_scores, 0.5), norm),
                old_div(percentile(final_scores, 0.1), norm),
                old_div(percentile(final_scores, 0.9), norm),
              )
          else:
            summary_report[d][run_label(run, short=True)] = (
                percentile(final_scores, 0.5) + norm + 1.0,
                percentile(final_scores, 0.1) + norm + 1.0,
                percentile(final_scores, 0.9) + norm + 1.0,
              )


    with open(self.args.stats_dir+ "/summary.dat", 'w') as o:
      # make summary report
      keys = sorted(reduce(set.union,
                           [set(x.keys()) for x in list(summary_report.values())],
                           set()))
      print('#####', end=' ', file=o)
      for k in keys:
        print(k, end=' ', file=o)
      print(file=o)
      for d, label_vals in sorted(summary_report.items()):
        print(d.split('/')[-2], end=' ', file=o)
        for k in keys:
          if k in label_vals:
            print('-', label_vals[k][0], label_vals[k][1], label_vals[k][2], end=' ', file=o)
          else:
            print('-', '-', '-', '-', end=' ', file=o)
        print(file=o)

    if keys:
      plotcmd = ["""1 w lines lt 1 lc rgb "black" notitle""",
                 """'summary.dat' using 3:4:5:xtic(1) ti "%s" """ % keys[0]]
      for n, k in enumerate(keys[1:]):
        plotcmd.append("""'' using %d:%d:%d ti "%s" """ % (
                        4*n + 7,
                        4*n + 8,
                        4*n + 9,
                        k))
      self.gnuplot_summary_file(self.args.stats_dir, 'summary', plotcmd)



    for d, label_runs in list(dir_label_runs.items()):
      labels = [k for k,v in list(label_runs.items())
                if len(v)>=self.args.min_runs]
      self.gnuplot_file(d,
                        "medianperfe",
                        ['"%s_percentiles.dat" using 1:12:4:18 with errorbars title "%s"' % (l,l) for l in labels])
      self.gnuplot_file(d,
                        "meanperfe",
                        ['"%s_percentiles.dat" using 1:21:4:18 with errorbars title "%s"' % (l,l) for l in labels])
      self.gnuplot_file(d,
                        "medianperfl",
                        ['"%s_percentiles.dat" using 1:12 with lines title "%s"' % (l,l) for l in labels])
      self.gnuplot_file(d,
                        "meanperfl",
                        ['"%s_percentiles.dat" using 1:21 with lines title "%s"' % (l,l) for l in labels])

    # print
    # print "10% Scores", d
    # pprint(self.technique_scores(d, labels, '0.1'))
    # print
    # print "90% Scores", d
    # pprint(self.technique_scores(d, labels, '0.9'))
    # print
    # print "Mean Scores", d
    # pprint(self.technique_scores(d, labels, 'mean'))
      print()
      print("Median Scores", d)
      pprint(self.technique_scores(d, labels, '0.5'))


  def technique_scores(self, directory, labels, ykey, xkey='#sec', factor=10.0):
    max_duration = None
    min_value = float('inf')
    for label in labels:
      try:
        dr = csv.DictReader(open(os.path.join(directory,label+"_percentiles.dat")), delimiter=' ', lineterminator='\n')
        lastrow = list(dr)[-1]
        max_duration = max(max_duration, float(lastrow[xkey]))
        min_value = min(min_value, float(lastrow[ykey]))
      except:
        log.exception("failed computing score")

    scores = list()

    for label in labels:
      try:
        dr = csv.DictReader(open(os.path.join(directory,label+"_percentiles.dat")), delimiter=' ', lineterminator='\n')
        score = 0.0
        lastsec = 0.0
        value = float('inf')
        for row in dr:
          duration = float(row[xkey]) - lastsec
          lastsec = float(row[xkey])
          value = float(row[ykey])
          score += duration * (value - min_value)
        score += (factor*max_duration - lastsec) * (value - min_value)
        scores.append((score, label))
      except:
        log.exception("failed computing score")

    return sorted(scores)


  def combined_stats_over_time(self,
                               output_dir,
                               label,
                               runs,
                               objective,
                               worst,
                               best,
                               ):
    """
    combine stats_over_time() vectors for multiple runs
    """

    #extract_fn = lambda dr: objective.stats_quality_score(dr.result, worst, best)
    extract_fn = _.result.time
    combine_fn = min
    no_data = 999

    log.debug("writing stats for %s to %s", label, output_dir)
    by_run = [self.stats_over_time(session, run, extract_fn, combine_fn, no_data)
              for run, session in runs]
    max_len = max(list(map(len, by_run)))

    by_run_streams = [Stream() << x << repeat(x[-1], max_len-len(x))
                      for x in by_run]
    by_quanta = list(zip(*by_run_streams[:]))

    def data_file(suffix, headers, value_function):
      with open(os.path.join(output_dir, label+suffix), 'w') as fd:
        out = csv.writer(fd, delimiter=' ', lineterminator='\n')
        out.writerow(['#sec'] + headers)
        for quanta, values in enumerate(by_quanta):
          sec = quanta*self.args.stats_quanta
          out.writerow([sec] + value_function(values))

   #data_file('_details.dat',
   #          map(lambda x: 'run%d'%x, xrange(max_len)),
   #          list)
   #self.gnuplot_file(output_dir,
   #                  label+'_details',
   #                  [('"'+label+'_details.dat"'
   #                    ' using 1:%d'%i +
   #                    ' with lines'
   #                    ' title "Run %d"'%i)
   #                   for i in xrange(max_len)])

    data_file('_mean.dat',
              ['#sec', 'mean', 'stddev'],
              lambda values: [mean(values), stddev(values)])
    self.gnuplot_file(output_dir,
                      label+'_mean',
                      ['"'+label+'_mean.dat" using 1:2 with lines title "Mean"'])

    def extract_percentiles(values):
      values = sorted(values)
      return ([values[int(round(p*(len(values)-1)))] for p in PCTSTEPS]
             + [mean(values)])
    data_file("_percentiles.dat", PCTSTEPS + ['mean'], extract_percentiles)
    self.gnuplot_file(output_dir,
                      label+'_percentiles',
                      reversed([
                        '"'+label+'_percentiles.dat" using 1:2  with lines title "0%"',
                      # '""                          using 1:3  with lines title "5%"',
                        '""                          using 1:4  with lines title "10%"',
                      # '""                          using 1:5  with lines title "25%"',
                        '""                          using 1:6  with lines title "20%"',
                      # '""                          using 1:7  with lines title "35%"',
                        '""                          using 1:8  with lines title "30%"',
                      # '""                          using 1:9  with lines title "45%"',
                        '""                          using 1:10 with lines title "40%"',
                      # '""                          using 1:11 with lines title "55%"',
                        '""                          using 1:12 with lines title "50%"',
                      # '""                          using 1:13 with lines title "65%"',
                        '""                          using 1:14 with lines title "70%"',
                      # '""                          using 1:15 with lines title "75%"',
                        '""                          using 1:16 with lines title "80%"',
                      # '""                          using 1:17 with lines title "85%"',
                        '""                          using 1:18 with lines title "90%"',
                      # '""                          using 1:19 with lines title "95%"',
                        '"'+label+'_percentiles.dat" using 1:20 with lines title "100%"',
                       ]))

  def gnuplot_file(self, output_dir, prefix, plotcmd):
    with open(os.path.join(output_dir, prefix+'.gnuplot'), 'w') as fd:
      print('set terminal postscript eps enhanced color', file=fd)
      print('set output "%s"' % (prefix+'.eps'), file=fd)
      print('set ylabel "Execution Time (seconds)"', file=fd)
      print('set xlabel "Autotuning Time (seconds)"', file=fd)
      print('plot', ',\\\n'.join(plotcmd), file=fd)

    try:
      subprocess.call(['gnuplot', prefix+'.gnuplot'], cwd=output_dir, stdin=None)
    except OSError:
      log.error("command gnuplot not found")

  def gnuplot_summary_file(self, output_dir, prefix, plotcmd):
    with open(os.path.join(output_dir, prefix+'.gnuplot'), 'w') as fd:
      print('set terminal postscript eps enhanced color', file=fd)
      print('set output "%s"' % (prefix+'.eps'), file=fd)
      print('''
set boxwidth 0.9
set style fill solid 1.00 border 0
set style histogram errorbars gap 2 lw 1
set style data histograms
set xtics rotate by -45
set bars 0.5
set yrange [0:20]

set yrange [0:10]
set key out vert top left
set size 1.5,1
set ytics 1

''', file=fd)
      print('plot', ',\\\n'.join(plotcmd), file=fd)
    subprocess.call(['gnuplot', prefix+'.gnuplot'], cwd=output_dir, stdin=None)


  def stats_over_time(self,
                      session,
                      run,
                      extract_fn,
                      combine_fn,
                      no_data = None):
    """
    return reduce(combine_fn, map(extract_fn, data)) for each quanta of the
    tuning run
    """
    value_by_quanta = [ no_data ]
    start_date = run.start_date

    subq = (session.query(Result.id)
           .filter_by(tuning_run = run, was_new_best = True, state='OK'))

    q = (session.query(DesiredResult)
         .join(Result)
         .filter(DesiredResult.state=='COMPLETE',
                 DesiredResult.tuning_run == run,
                 DesiredResult.result_id.in_(subq.subquery()))
         .order_by(DesiredResult.request_date))

    first_id = None
    for dr in q:
      if first_id is None:
        first_id = dr.id
      td = (dr.request_date - start_date)
      duration = td.seconds + (td.days * 24 * 3600.0)
      if self.args.by_request_count:
        quanta = dr.id - first_id
      else:
        quanta = int(old_div(duration, self.args.stats_quanta))
      while len(value_by_quanta) <= quanta:
        value_by_quanta.append(value_by_quanta[-1])

      if value_by_quanta[-1] is no_data:
        value_by_quanta[-1] = extract_fn(dr)
      else:
        value_by_quanta[-1] = combine_fn(value_by_quanta[-1], extract_fn(dr))

    return value_by_quanta





if __name__ == '__main__':
  opentuner.tuningrunmain.init_logging()
  sys.exit(StatsMain(argparser.parse_args()).main())


